#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: common.py 666 2009-09-14 16:01:03Z smrcka $
'default settings and some common functions'

from time import time
import model as m

#=============================================================================
# global options
OPTS = {
    'verbose': 1,
    'filename': '',
    'noprint': 0,       # do not print the extended model
    'info': 0,
    'ofile': 'extmodel.smv',
    'methods': {},
    'clk_enable': [],   # clock enable signals
    'clk_val': [],      # clock value signals
    'auto_clk': 0,      # auto mode when identifying clk enable and value signals
}

#=============================================================================
# debugging functions
def debug(lvl, text):
    "print the debug info, if OPTS['verbose'] is >= lvl"
    if OPTS['verbose'] >= lvl:
        for line in str(text).split(u'\n'):
            print line
#            print '-- %s' % line

def debug1(text):
    "print the debug info, if OPTS['verbose'] is >=1"
    debug(1, text)

def debug2(text):
    "print the debug info, if OPTS['verbose'] is >=2"
    debug(2, text)

def debug3(text):
    "print the debug info, if OPTS['verbose'] is >=3"
    debug(3, text)

def debug4(text):
    "print the debug info, if OPTS['verbose'] is >=4"
    debug(4, text)

#=============================================================================
# logging function
def prelog(f):
    def wrapper(*args, **kargs):
        print '\n--DEBUG-- called function: %s' % f.__name__
        if args:
            print '--DEBUG-- args: %r' % args
        if kargs:
            print '--DEBUG-- kargs: %r' % kargs
        return f(*args, **kargs)
    return wrapper

log = prelog

def postlog(f):
    def wrapper(*args, **kargs):
        now = time()
        try:
            return f(*args, **kargs)
        finally:
            print '\n--DEBUG-- called function: %s' % f.__name__
            if args:
                print '--DEBUG-- args: %r' % args
            if kargs:
                print '--DEBUG-- kargs: %r' % kargs
            print "--DEBUG-- time delta: %s" % (time() - now)
    return wrapper

#=============================================================================
# model operations 
EDGE_T = {
    'rising': '/',
    'falling': "\\",
#    'rising': u'↑',
#    'falling': u'↓',
}

def clk2str(clock):
    'returns string representation of clock'
    var, edge = clock
    return EDGE_T[edge] + var.name

def find_clocks(expr):
    """return the list of tuples (variable, edge_type), in which the expr
    contains clock edge detection. The function is able to detect the edge
    expression of type: (next(c) & ~c) or (~c & next(c)).
    edge_type \in {'rising', 'falling'}
    Called by module_clocks()."""
    def get_all_and_operands(expr_list):
        "returns all 'and' operands, regardless the depth of the sub-and-expr"
        # ((a & ~b) & next(b))  ===>  (a & ~b & next(b))
        # returns list: [a, ~b, next(b)]
        and_operands = []
        for expr in expr_list:
            if isinstance(expr, m.Operator) and expr.operator == '&':
                and_operands += get_all_and_operands(expr.expr_list)
            else:
                and_operands.append(expr)
        return and_operands
        
    # the resulting list of tuples
    sensitive = []

    # at first, look at the expression if the clock sensititivy isn't here...
    # if the expression is AND operator, the clock sensitivity might be in the
    # 'expr'
    if isinstance(expr, m.Operator) and expr.operator == '&':
        # for a simplicity of the detection, we need to flatten the expression
        # list of operator expression, i.e.,
        expr_list = get_all_and_operands(expr.expr_list)

        # look for rising edge... (~c & next(c))
        # forall node1,node2 in expr.expr_list, node1 != node2:
        for node1 in expr_list:
            for node2 in [e for e in expr_list if e != node1]:
                # if  node1 == ~c
                # and node2 == next(c)
                if isinstance(node1, m.UnaryOp) and node1.operator == '~' and \
                  isinstance(node1.expr, m.Variable) and \
                  isinstance(node2, m.NextOp) and node1.expr == node2.expr:
                    clock = (node1.expr, 'rising')
                    if clock not in sensitive:
                        sensitive.append(clock)

        # look for falling edge... (c & ~next(c))
        # forall node1,node2 in expr_list, node1 != node2:
        for node1 in expr_list:
            for node2 in [e for e in expr_list if e != node1]:
                # if  node1 == c
                # and node2 == ~next(c)
                if isinstance(node1, m.Variable) and \
                  isinstance(node2, m.UnaryOp) and node2.operator == '~' and \
                  isinstance(node2.expr, m.NextOp) and \
                  node2.expr.expr == node1:
                    clock = (node1, 'falling')
                    if clock not in sensitive:
                        sensitive.append(clock)

    # then, process subnodes of the expression
    if isinstance(expr, m.UnaryOp):
        sensitive += find_clocks(expr.expr)

    elif isinstance(expr, m.Operator):
        for operand in expr.expr_list:
            # we must ADD other clock sensitivity in the expr if we already
            # found some by checking the AND operator expression (above)
            sensitive += find_clocks(operand)

    # similarly for other types of an expression
    elif isinstance(expr, m.TernaryOp):
        for operand in expr.data:
            sensitive += find_clocks(operand)

    elif isinstance(expr, m.ArrayCat):
        for operand in expr.data:
            sensitive += find_clocks(operand)

    elif isinstance(expr, m.TempOp):
        sensitive += find_clocks(expr.expr)

    elif isinstance(expr, m.SwitchOp):
        sensitive += find_clocks(expr.cond)
        for con1, con2 in expr.cases:
            sensitive += find_clocks(con1)
            sensitive += find_clocks(con2)

    elif isinstance(expr, m.CaseBlock):
        for con1, con2 in expr.cases:
            sensitive += find_clocks(con1)
            sensitive += find_clocks(con2)

    return sensitive

def check_module_clocks(modinst):
    """returns clocks (tuples: (variable,edge_type)) of module instance
    statement.
    1. find module clocks via module_clocks
    2. check if all the clocks are input parameters of instantiated module
    3. check if the clock parameter is connected to the variable/parameter (not
       some kind of expression)
    4. transform clock parameter of instantiated module to the connected
       variable and return changed clock:
        (variable_connected_to_module, edge_type)
    Raise TypeError if any of conditions is violated.
    Called by module_clocks()."""
    # find module clocks via module_clocks
    modsens_list = module_clocks(modinst.module)

    # check if all the clocks are input parameters of instantiated module
    # substitute every variable in modsens_list with parameter if it is an
    # alias

    updated_sens_list = []
    for var, edge in modsens_list:
        var1 = modinst.module.get_param_from_alias(var)
        if not(isinstance(var1, m.ModuleParameter) and \
           var1.direction == 'input'):
            raise TypeError( \
                "clock signal '%s' in module '%s: %s' is not an input port" % \
                (var.name, modinst.name, modinst.module.name))
#        if var != var1:
#            print 'substituting %s with %s' % (var, var1)
        updated_sens_list.append((var1, edge))
    modsens_list = updated_sens_list
#    print ','.join('(%s, %s)' % (str(v),e) for v,e in modsens_list)

    # check if the clock parameter is connected to the variable/parameter (not
    # some kind of expression)
    for var, edge in modsens_list:
        # get the expression connected to the clock port
        clk_expr = modinst.param_map[var.name]
        if not isinstance(clk_expr, m.Variable):
            raise TypeError("At module instantiation: %s\n" \
                "unsupported expression at clock port '%s'\n" \
                "variable or parameter expected" % (modinst, var))

    # transform clock parameter of instantiated module to the connected
    # variable and return changed clock
    return [(modinst.param_map[var.name], edge) \
        for var, edge in modsens_list]

def module_clocks(module):
    """find and return list of clock signals - either module parameter names or
    module local variable names. clock sensitivity list will be set in
    module.clocks and every module.clocks of instantiated submodules.
    Called recursively via check_module_clocks() and by extend()."""

    # inspect every module statement and every its expression for find_clocks
    sens_list = []

    # in every module statement
    for stmt in module.stmts:

        # find expressions accessed as read
        for expr in stmt.read_expr:
            # and in every expression find possible clock detection
            sens_list += find_clocks(expr)

        # moreover, if the statement is module instance statement, we need to
        # inspect the module itself for clock detection too
        if isinstance(stmt, m.ModuleInstStmt):
            sens_list += check_module_clocks(stmt)

    # clock signal map pointing to representative of a clock variable
    module.clk_sig_map = {}

    # try to substitute every variable with its parameter alias, take the first
    # found module parameter in aliases list
    clock_sens_list = []
    for clock, edge in sens_list:
        params = [varx for var1, varx in module.aliases \
            if var1 == clock and isinstance(varx, m.ModuleParameter)]
        if params:
            clock_sens_list.append((params[0], edge))
            module.clk_sig_map[clock] = params[0]
        else:
            clock_sens_list.append((clock, edge))

    # substitute variable clocks with their first alias representative
    sens_list = []
    for clock, edge in clock_sens_list:
        #if isinstance(clock, m.Variable): <--EE!should be exactly the Variable
        if type(clock) == m.Variable:
            a_class = [clock]
            for a_class in [c for c in module.alias_classes if clock in c]:
                break
            sens_list.append((a_class[0], edge))
            module.clk_sig_map[clock] = a_class[0]
        else:
            sens_list.append((clock, edge))
            module.clk_sig_map[clock] = clock
    
    module.clocks = uniq_list(sens_list)
    return module.clocks

def uniq_list(list1):
    'remove duplicated items in the list'
    result = []
    for item in list1:
        if item not in result:
            result.append(item)
    return result

UNIQ_ID = 0

def uniqname(name, delimiter = 'x'):
    '''returns uniq name wrt name and UNIQ_ID
    used by flatten_module for renaming local variables'''
    global UNIQ_ID
    UNIQ_ID += 1
    return '%s_%s%s' % (name, delimiter, UNIQ_ID)

def print_found_clocks(model, lvl=None):
    'prints which modules are asynchronous, synchronous, and transparent'
    if lvl == None:
        cond = OPTS['info']
    else:
        cond = OPTS['verbose'] >= lvl
    if cond: 
        print('-- %-30s | detected clock (%s = rising, %s = falling)' % \
            ('Module', EDGE_T['rising'], EDGE_T['falling']))
        print('-'*78)
        print('-- Asynchronous modules:')
        for module in model.modules:
            if len(module.clocks) > 1:
                print('-- %-30s | %s' % (module.name, \
                    u', '.join([clk2str(clock) for clock in module.clocks])))

        print('-- Synchronous modules:')
        for module in model.modules:
            if len(module.clocks) == 1:
                print('-- %-30s | %s' % (module.name, \
                    ', '.join([clk2str(clock) for clock in module.clocks])))

        print('-- Transparent modules:')
        for module in model.modules:
            if len(module.clocks) == 0:
                print('-- %-30s |' % module.name)
        print

def find_parent_modules(startmod, module):
    'returns list of modules which instantiate the module'
    mods = []
    for stmt in startmod.stmts:
        if isinstance(stmt, m.ModuleInstStmt):
            if stmt.module == module:
                mods.append(startmod)
            mods += find_parent_modules(stmt.module, module)
    return uniq_list(mods)

def flatten_module(supmodule, modinst):
    '''changes supmodule such that module instance modinst (which is supmodule
    statement) will be flattened'''
    prestmts = supmodule.stmts[:supmodule.stmts.index(modinst)]
    poststmts = supmodule.stmts[supmodule.stmts.index(modinst)+1:]
    fstmts = []

    # 1st copy just for making able to change module variables' names
    module_copy = modinst.module.mkinstance()
    var_map = modinst.param_map.copy()
    for paramname, var in modinst.param_map.items():
        if isinstance(var.type, m.ArrayType):
            for par_item, expr_item in \
              zip(modinst.module.parameters[paramname].data, var.data):
                var_map[par_item.name] = expr_item
    for varname, var in module_copy.variables.items():
        nvar = var.copy()
        uqname = uniqname(varname)
        nvar.name = uqname          # hope it will also change smv_name
        var_map[varname] = nvar
        if isinstance(var.type, m.ArrayType):
            old_names = [i.name for i in nvar.data]
            nvar.rename_data_items(uqname)
            for item_name, item in zip(old_names, nvar.data):
                var_map[item_name] = item
        # and add a new variable reference to .variables of superior module
        supmodule.variables[nvar.name] = nvar

    # 2nd copy with a proper substitution of expressions
    for stmt in module_copy.stmts:
        if isinstance(stmt, m.ModParamDeclStmt):
            pass
        else:
            nstmt = m.copy_stmt(stmt, var_map)
            if isinstance(nstmt, m.ModuleInstStmt) or \
              isinstance(nstmt, m.AssertionStmt):
                nstmt.name = uniqname(nstmt.name)
            fstmts.append(m.copy_stmt(stmt, var_map))

    supmodule.stmts = prestmts + fstmts + poststmts

    supmodule.find_aliases()

def flatten_all_asyn_modules(model):
    '''finds asynchronous modules and flatten all of them except only one,
    possibly, in the hiearchy, at the top of the rest'''
    # mark used modules
    model.set_used_modules()
    # find and set the clock sensitivity list in every module
    module_clocks(model.main_module)
    # find asynchronous modules
    asyn_modules = [mod for mod in model.modules if len(mod.clocks) > 1]
    # print the info
    print_found_clocks(model, 4)

    while len(asyn_modules) > 1:
        # find parents of module to be flatten...
        what_flatten = asyn_modules[-1]
        debug2('Flattenning asyn. module: %s' % what_flatten.name)
        where_flatten = find_parent_modules(model.main_module, what_flatten)
        for parent in where_flatten:
            for stmt in parent.stmts:
                if isinstance(stmt, m.ModuleInstStmt) and \
                  stmt.module == what_flatten:
                    # ...and flatten it
                    flatten_module(parent, stmt)
    
        # mark used modules
        model.set_used_modules()
        # find and set the clock sensitivity list in every module
        module_clocks(model.main_module)
        # find asynchronous modules
        asyn_modules = [mod for mod in model.modules if len(mod.clocks) > 1]
        # print the info
        print_found_clocks(model, 4)

    print_found_clocks(model)

def flatten_renaming_modules(model):
    "Flatten every renaming module instance."
    debug2('Flat modules (probably input/output ports of the component)...')
    debug2('%40s | %s' % ('module instance', 'module'))
    debug2('-'*76)
    flatten_renaming_module(model.main_module)

def flatten_renaming_module(module):
    """Flatten every renaming module instance.
    Renaming module is a module which includes the simplest assignment
    statements or renaming module instances only.
    The simplest assignment statement is a statement:
      variable1 := variable2;   or
      variable := constant;
    Returns True if the module itself is renaming module.
    Called recursivelly and by extend()."""
    # e.g.,
    # module gnd(p) {
    #   output p: boolean;
    #   p := 1;
    # }
    #
    # or
    #
    # module ibuf(i, o) {
    #   input i: boolean;
    #   output o: boolean;
    #   o := i;
    # }

    # flatten all renaming submodules
    closed = []
    flat = True
    while flat:
        flat = False
        for stmt in module.stmts:
            if stmt not in closed:
                if isinstance(stmt, m.ModuleInstStmt) and \
                  flatten_renaming_module(stmt.module):
                    debug2('%40s | %s' % 
#                    debug2( \
#                        'Flat module instance: %s : %s' % \
                        (stmt.name, stmt.module.name))
                    flatten_module(module, stmt)
                    flat = True
                    break
                closed.append(stmt)

    module.find_aliases()

    # get if the module is renaiming module
    is_renaming = True
    for stmt in module.stmts:
        # ignore declaration statements
        if isinstance(stmt, m.VariableDeclStmt) or \
          isinstance(stmt, m.ModParamDeclStmt):
            continue
        # try to hold the renaming module condition
        is_renaming = is_renaming and \
          isinstance(stmt, m.AssignmentStmt) and \
          isinstance(stmt.dst, m.Variable) and \
          (isinstance(stmt.src, m.Variable) or \
           isinstance(stmt.src, m.Constant))
        # do not continue if found not-renaming module
        if not is_renaming:
            break

    return is_renaming

class NotCombLogic(Exception):
    'not combinational logic expression exception'
    pass

def comb_logic_to_structural(model):
    '''Within the main module, replace every assignment statement with
    combinational logic with its structural description using and, or, and not
    modules.
    Note: combinational logic description should be at the top level of
    indentation of module statements (not within if-then-else or other nested
    statements.'''
    def create_logic_modules():
        'returns logic modules wrt. syntax tree data structure'
        # AND module
        and_mod = m.Module('and', ['o', 'i1', 'i2'])
        o_param = m.ModuleParameter('o', m.BOOLEAN, 'output')
        i1_param = m.ModuleParameter('i1', m.BOOLEAN, 'input')
        i2_param = m.ModuleParameter('i2', m.BOOLEAN, 'input')
        and_mod.parameters = { 'o': o_param, 'i1': i1_param, 'i2': i2_param }
        and_mod.stmts = [ \
            m.ModParamDeclStmt(o_param), \
            m.ModParamDeclStmt(i1_param), \
            m.ModParamDeclStmt(i2_param), \
            m.AssignmentStmt(o_param, \
                m.Operator('&', [i1_param, i2_param], False), \
                False) \
            ]
        and_mod.clocks = []
        # OR module
        or_mod = m.Module('or', ['o', 'i1', 'i2'])
        o_param = m.ModuleParameter('o', m.BOOLEAN, 'output')
        i1_param = m.ModuleParameter('i1', m.BOOLEAN, 'input')
        i2_param = m.ModuleParameter('i2', m.BOOLEAN, 'input')
        or_mod.parameters = { 'o': o_param, 'i1': i1_param, 'i2': i2_param }
        or_mod.stmts = [ \
            m.ModParamDeclStmt(o_param), \
            m.ModParamDeclStmt(i1_param), \
            m.ModParamDeclStmt(i2_param), \
            m.AssignmentStmt(o_param, \
                m.Operator('|', [i1_param, i2_param], False), \
                False) \
            ]
        or_mod.clocks = []
        # NOT module
        not_mod = m.Module('not', ['o', 'i'])
        o_param = m.ModuleParameter('o', m.BOOLEAN, 'output')
        i_param = m.ModuleParameter('i', m.BOOLEAN, 'input')
        not_mod.parameters = { 'o': o_param, 'i': i_param }
        not_mod.stmts = [ \
            m.ModParamDeclStmt(o_param), \
            m.ModParamDeclStmt(i_param), \
            m.AssignmentStmt(o_param, m.UnaryOp('~', i_param, False), False) \
            ]
        not_mod.clocks = []
        # return all
        return (and_mod, or_mod, not_mod)

    def is_comb_logic_stmt(stmt):
        'returns True if stmt is combinational logic assignemnt statement'
        return isinstance(stmt, m.AssignmentStmt) and \
           stmt.dst not in excp and \
          (isinstance(stmt.src, m.Operator) or \
           isinstance(stmt.src, m.UnaryOp)) and \
          stmt.src.operator in ('&', '|', '~')

    def structural_description(output, expr):
        '''returns statement list of structural description of assinment of
        expr to output'''
        stmts = []
        vars = {}
        if isinstance(expr, m.Operator) and expr.operator in ('&', '|'):
            operator = 'and' if expr.operator == '&' else 'or'
            assert len(expr.expr_list) == 2
            # left operand transformation
            oprnd = expr.expr_list[0]
            if isinstance(oprnd, m.Variable) or isinstance(oprnd, m.Constant):
                left = oprnd
            else:
                left = m.Variable(uniqname('comb'), m.BOOLEAN, False)
                vars[left.name] = left
                stmts.append(m.VariableDeclStmt(left))
                new_stmts, new_vars = structural_description(left, oprnd)
                stmts += new_stmts
                vars.update(new_vars)
            # right operand transformation
            oprnd = expr.expr_list[1]
            if isinstance(oprnd, m.Variable) or isinstance(oprnd, m.Constant):
                right = oprnd
            else:
                right = m.Variable(uniqname('comb'), m.BOOLEAN, False)
                vars[right.name] = right
                stmts.append(m.VariableDeclStmt(right))
                new_stmts, new_vars = structural_description(right, oprnd)
                stmts += new_stmts
                vars.update(new_vars)
            # operator transformation
            #   insert new module to model if it does not exist
            if not modinst_number[operator]:
                mod = and_mod if operator == 'and' else or_mod
                model.modules.append(mod)
                model.stmts.append(mod)
            modinst_number[operator] += 1

            mod_inst = m.ModuleInstStmt(uniqname(operator, 'g'), \
                operator, \
                [output, left, right] \
                )
            mod_inst.unify(model)
            stmts.append(mod_inst)
        elif isinstance(expr, m.UnaryOp) and expr.operator == '~':
            # operand transformation
            if not modinst_number['not']:
                model.modules.append(not_mod)
                model.stmts.append(not_mod)
            modinst_number['not'] += 1

            oprnd = expr.expr
            if isinstance(oprnd, m.Variable) or isinstance(oprnd, m.Constant):
                left = oprnd
            else:
                left = m.Variable(uniqname('comb'), m.BOOLEAN, False)
                vars[left.name] = left
                stmts.append(m.VariableDeclStmt(left))
                new_stmts, new_vars = structural_description(left, oprnd)
                stmts += new_stmts
                vars.update(new_vars)
            # operator transformation
            mod_inst = m.ModuleInstStmt(uniqname('not', 'g'),\
                'not',
                [output, left])
            mod_inst.unify(model)
            stmts.append(mod_inst)
        elif set(get_vars(expr)).intersection(excp):
            raise NotCombLogic()
        else:
            raise NotImplementedError(
                "%s:%d: unsupported combinational logic expression `%s'" % \
                (OPTS['filename'], expr.lineno, expr))
        return (stmts, vars)
    
    # make exceptions for transformation (clock signals and its predecessors)
    excp = [c for c, edge in model.main_module.clocks]
    preds = []
    for stmt in model.main_module.stmts:
        if isinstance(stmt, m.AssignmentStmt) and stmt.dst in excp:
            preds += get_vars(stmt.src)
    excp += preds
    # add comb. logic modules
    and_mod, or_mod, not_mod = create_logic_modules()
    modinst_number = {'and': 0, 'or': 0, 'not': 0}
    # create new module statement list
    stmts = []
    for stmt in model.main_module.stmts:
        if is_comb_logic_stmt(stmt):
            if isinstance(stmt.dst, m.TempOp):
                stmts.append(stmt)
                continue
            try:
                new_stmts, new_vars = structural_description(
                    stmt.dst, stmt.src)
            except NotCombLogic:
                stmts.append(stmt)
                continue
            debug4('Transforming statement:')
            debug4('  %s' % stmt)
            debug4('Got this:')
            for s in new_stmts:
                debug4(s)
            stmts.append(m.CommentStmt(stmt))
            stmts += new_stmts
            model.main_module.variables.update(new_vars)
        else:
            stmts.append(stmt)
    model.main_module.stmts = stmts

def find_module_to_extend(model):
    '''returns a module to extend wrt. its asynchronicity and
    OPTS['extend'].
    returns None if error'''
    asyn_modules = [mod for mod in model.modules if len(mod.clocks) > 1]
    if not asyn_modules:
        print('-- Nothing to extend')
        if not OPTS['info']:
            print('Try -s for more details')
        return None
    else:
        return asyn_modules[0]

def partition_clock_domain(model):
    '''partition all synchronous module instances to clock domains,
    model.clock_domains: (variable, edge_type) -> [module instances]
    module_instance.clock_domain: (variable, edge_type) or None'''
    for stmt in model.asyn_module.stmts:
        if isinstance(stmt, m.ModuleInstStmt):
            # is a module synchronous?
            if stmt.module.clocks:
                assert len(stmt.module.clocks) == 1, \
                    'clock sensitivity list should contain one item only'
                # which clock signal?
                # (the-only-one clock, variable)    v  v
                clk_param_name = stmt.module.clocks[0][0].name
                clk_expr = stmt.param_map[clk_param_name]
                assert isinstance(clk_expr, m.Variable), \
                    'variable or parameter must be connected to clock port: %s'
                clock = (model.asyn_module.clk_sig_map[clk_expr],
                         stmt.module.clocks[0][1])
                if clock not in model.clock_domains:
                    model.clock_domains[clock] = []
                model.clock_domains[clock].append(stmt)
                # mark clock domain within the module
                stmt.clock_domain = clock
            else:
                stmt.clock_domain = None

def print_clock_domains(model):
    'prints module instace with clock signal on which it is sensitive'
#    if OPTS['verbose'] >= 4:
    debug3('%-20s | %-20s | %s' % \
        ('clock signal', 'module', 'instance name'))
    debug3('-'*75)
#    for clock in model.clock_domains:
#        print '------------------------------------'
#        print model.clock_domains[clock]
    for clock, inst in model.clock_domains.items():
        for i in inst:
            debug3('%-20s | %-20s | %s' % \
                (clk2str(clock), i.module.name, i.name))

def get_vars(expr):
    'return a list of variables occuring in expression'
    if isinstance(expr, m.Variable):  
        return [expr]
    elif isinstance(expr, m.UnaryOp):
        return get_vars(expr.expr)
    elif isinstance(expr, m.Operator):
        return reduce(lambda x, y: x+y, [get_vars(e) for e in expr.expr_list])
    elif isinstance(expr, m.TernaryOp):
        return reduce(lambda x, y: x+y, [get_vars(e) for e in expr.data])
    elif isinstance(expr, m.ArrayCat):
        return reduce(lambda x, y: x+y, [get_vars(e) for e in expr.data])
    elif isinstance(expr, m.TempOp):
        return get_vars(expr.expr)
    elif isinstance(expr, m.SwitchOp):
        variables = get_vars(expr.cond)
        for expr1, expr2 in expr.cases:
            variables += get_vars(expr1)
            variables += get_vars(expr2)
        return variables
    elif isinstance(expr, m.CaseBlock):
        variables = []
        for expr1, expr2 in expr.cases:
            variables += get_vars(expr1)
            variables += get_vars(expr2)
        return variables
    else:
        return []

def get_ports(modinst, direction, with_clocks = True):
    '''returns list of variables connected to input/output parameters of module
    instance, direction is either "input" or "output"'''
    ports = []
    if with_clocks:
        exprs = [modinst.param_map[p.name] \
            for p in modinst.module.parameters.values() \
            if p.direction == direction]
    else:
        clock_parameters = [c[0] for c in modinst.module.clocks]
        exprs = [modinst.param_map[p.name] \
            for p in modinst.module.parameters.values() \
            if p.direction == direction and p not in clock_parameters]
    for expr in exprs:
        ports += get_vars(expr)
    return ports

def add_sens(sens, var, clock):
    '''Enhance sens (sensitivity map) with variable and a clock. If the
    variable is of an array type, enhance the sensitivity for all of its
    items'''
    if var in sens:
        sens[var].add(clock)
    else:
        sens[var] = set([clock])
    if isinstance(var.type, m.ArrayType):
        for item in var.data:
            add_sens(sens, item, clock)

def create_sens_map(model):
    '''Makes a map model.sens which will contain every model Variable or
    Parameter pointing to an empty set. It will also contain a mapping from
    items of bit arrays'''
    model.sens = {}   # Variable->set of clocks
    for var in model.asyn_module.variables.values():
        model.sens[var] = set()
        if isinstance(var.type, m.ArrayType):
            for item in var.data:
                model.sens[item] = set()
    for var in model.asyn_module.parameters.values():
        model.sens[var] = set()
        if isinstance(var.type, m.ArrayType):
            for item in var.data:
                model.sens[item] = set()

def copy_sens_map(sens):
    'return a copy of a sensitivity map'
    sens_copy = sens.copy()
    for key, value in sens_copy.items():
        sens_copy[key] = value.copy()
    return sens_copy

def add_output_port_sens(model):
    '''returns map Variable->set of clocks for variable connected to output
    of synchronous modules'''
    # find output ports and set the clock sensitivity to them
    for clock, mlist in model.clock_domains.items():
        # for every module instance in a clock domain
        for minst in mlist:
            # and for every variable connected to its ouput
            for var in get_ports(minst, 'output'):
                # adjust the sensitivity map
                add_sens(model.sens, var, clock)
                # also adjust the sensitivity map for every variable within the
                # same alias equivalence class
                for a_class in model.asyn_module.alias_classes:
                    if var in a_class:
                        for alias in a_class:
                            add_sens(model.sens, alias, clock)
                        break

def add_array_sens(model):
    'add sensitivity to every root variable of an array'
    for var in model.asyn_module.variables.values():
        if isinstance(var.type, m.ArrayType):
            for item in var.data:
                model.sens[var] = model.sens[var].union(model.sens[item])

def print_sens_map(model):
    'prints sensitivity map'
    if OPTS['verbose'] >= 4:
        print '-'*76
        print 'Variable sensitivity'
        items = model.sens.items()
        items.sort(lambda x, y: -1 if x[0].name < y[0].name else 1)
        print('%40s | %s' % ('variable', 'changing event'))
        print '-'*76
        for var, clocks in items:
            print('%40s | %s' % (var, ','.join([clk2str(c) for c in clocks]) if
            clocks else '(external)'))

def set_dep_sens(model):
    '''set the sensitivity on clock wrt. assignment statements and transparent
    module instances'''
    def get_rw_exprs(stmt):
        'returns (reading, writing) lists of variables'
        if isinstance(stmt, m.ModuleInstStmt):
            if not stmt.module.clocks:
                exprs = get_ports(stmt, 'input')
                reading = []
                for expr in exprs:
                    reading += get_vars(expr)
                exprs = get_ports(stmt, 'output')
                writing = []
                for expr in exprs:
                    writing += get_vars(expr)
                return (reading, writing)
            else:
                return ([], [])

        elif isinstance(stmt, m.AssignmentStmt):
            return (get_vars(stmt.src), get_vars(stmt.dst))

        elif isinstance(stmt, m.IfThenElseStmt):
            reading = get_vars(stmt.cond)
            re1, wr1 = get_rw_exprs(stmt.then_stmt)
            re2, wr2 = get_rw_exprs(stmt.else_stmt)
            return (reading + re1 + re2, wr1 + wr2)

        elif isinstance(stmt, m.DoStmt) or isinstance(stmt, m.StmtListStmt):
            reading = []
            writing = []
            for stmt1 in stmt.stmt_list:
                re1, wr1 = get_rw_exprs(stmt1)
                reading += re1
                writing += wr1
            return (reading, writing)

        else:
            return ([], [])
   
    # for every statement
    for stmt in model.asyn_module.stmts:
        # get the variables it is reading from/writing to
        reading, writing = get_rw_exprs(stmt)
        # add to every written variable clock sensitivity from its source
        for wr_var in writing:
            for rd_var in reading:
                model.sens[wr_var] = \
                    model.sens[wr_var].union(model.sens[rd_var])

def transp_depen_stmts(module):
    'returns the list of assign+modinst statements in the module'

    def get_stmts(stmt):
        'return the list of assign+modinst statements ignoring indentation'
        # only nontemporal assignment statements and transparent module
        # instances is interesting
        if isinstance(stmt, m.AssignmentStmt) and \
          not isinstance(stmt.dst, m.TempOp):
            return [stmt]
        elif isinstance(stmt, m.ModuleInstStmt) and \
          len(stmt.module.clocks) == 0:
            return [stmt]
        elif isinstance(stmt, m.IfThenElseStmt):
            return get_stmts(stmt.then_stmt) + get_stmts(stmt.else_stmt)
        elif isinstance(stmt, m.DoStmt) or isinstance(stmt, m.StmtListStmt):
            res = []
            for stm in stmt.stmt_list:
                res += get_stmts(stm)
            return res
        else:
            return []

    # get usable statements
    stmts = []
    for stmt in module.stmts:
        stmts += get_stmts(stmt)
    
    for stmt in stmts:
        assert isinstance(stmt, m.AssignmentStmt) or \
               isinstance(stmt, m.ModuleInstStmt)

    return stmts

def make_direct_dependency(model):
    'makes model.direct_dep : Variable->[Variable] of the asynchronous module'
    # get all assignment or module instance statements from the asyn. module
    model.direct_dep = {}
    for stmt in transp_depen_stmts(model.asyn_module):
        if isinstance(stmt, m.AssignmentStmt):
            input = get_vars(stmt.src)
            output = get_vars(stmt.dst)
            assert len(output) == 1
            assert output[0] not in model.direct_dep
            model.direct_dep[output[0]] = input
        else: # stmt is m.ModuleInstStmt
            input = get_ports(stmt, 'input')
            output = get_ports(stmt, 'output')
            for var in output:
                assert var not in model.direct_dep
                model.direct_dep[var] = input

def print_trans_dep(model, var):
    debug3('transp. dependency: %s <= %s' % (var, \
        ', '.join([v.name for v in model.trans_dep[var]])))
    debug3('')

def transp_depen(model, var):
    'returns set of synchronized variables on which var depends'
    # DFS search for all variables transparently setting (i.e., not being
    # sensed by any clock signal) the value of var

    ### !!! nekde tady je chyba; z open odejde promenna, promeni se na vic
    # promennych, ze kterych se casem zase stane jiz odejita - v dfs by to mela
    # kryt closed mnozina, jenze ta se tady zda nebyt pro tento ucel stvorena
    open = [var]
    closed = []
    synchronized = []
    i = 0
    br = var.name.endswith('emuxcyo_4_')
    while open:
        i += 1
#        debug3('------------ Iteration %d' % i)
#        debug3('open = %s' % [str(v) for v in open])
#        debug3('closed = %s' % [str(v) for v in closed])
#        debug3('synchronized = %s' % [str(v) for v in synchronized])
        current = open.pop()
        closed.append(current)
#        if br:
#            print '='*76
#            print "current = '%s'" % current.name
#            print 'open = %s' % [v.name for v in open]
#            print 'closed = %s' % [v.name for v in closed]
#            print 'sync = %s' % [v.name for v in synchronized]
        if model.sens[current]:
            synchronized.append(current)
            continue
        if current in model.trans_dep:
#            if br:
#                print "Using trans_dep['%s'] = %s" % (current.name,
#                    [v.name for v in model.trans_dep[current]])
            for trans_var in model.trans_dep[current]:
                if (trans_var not in open) and (trans_var not in closed):
                    open.append(trans_var)
            continue
        # find all inputs which set the output 'current' wrt. assignment
        # statement or transparent module instance
        try:
#            if br:
#                print 'Direct dependency: %s' % [v.name for v in model.direct_dep[current]]
            for item in model.direct_dep[current]:
                if (item not in open) and (item not in closed):
                    open.append(item)
        except KeyError:
            pass
#        predecessors = []
#        for stmt in stmts:  
#            # either assignemnt statement
#            if isinstance(stmt, m.AssignmentStmt) and \
#              current in get_vars(stmt.dst):
#                predecessors += get_vars(stmt.src)
#            # or module instance
#            elif isinstance(stmt, m.ModuleInstStmt) and \
#              current in get_ports(stmt, 'output'):
#                predecessors += get_ports(stmt, 'input')
#        for item in predecessors:
#            if (item not in open) and (item not in closed):
#                open.append(item)
    return synchronized

def create_transparent_deps(model):
    '''makes model.trans_dep being the map of Variable->[Variable]
    creates dependencies from transparent variables to their synchronized 
    dependency predecessors
    useful to get the input of a signal path'''
    make_direct_dependency(model)
    model.trans_dep = {}
    # for every variable in the module
    for var in model.asyn_module.variables.values() + \
               model.asyn_module.parameters.values():
        # if the variable is in the sensitivity map
        if model.sens[var]:
            # make it dependent on itself
            model.trans_dep[var] = [var]
        else:
            # or find all its inputs and mark the depencency on them
            deps = transp_depen(model, var)
            model.trans_dep[var] = deps if deps else [var]

def clockset2str(clockset):
    return ','.join([clk2str(c) for c in clockset])

def get_modinst_list(model):
    'returns list of module instance statement of the asynchronous module'
    return [stmt for stmt in model.asyn_module.stmts \
        if isinstance(stmt, m.ModuleInstStmt)]

def make_src_dst_flow(model):
    '''1. makes src and dst data flow
    2. mark init clocks for every flow
    model.src_flow: Var1->([Var2],set(clock)); Var2 are setting value of Var1
    Wire = (gate, signal)
    model.dst_flow: Wire1->([Wire2],set(clock)) that Wire1 influences Wire2'''
    # initialize flows
    src_flow = {}
    dst_flow = {}
    iterable_src = []
    iterable_dst = []
    signal_to_gates = {}
    # 1. for every gate...
    debug4('\nSignals over which calculate dst transitive closure')
    for gate in model.gates:
        inputs = uniq_list(get_ports(gate, 'input', False))
        outputs = uniq_list(get_ports(gate, 'output', False))
        # ...and every tuple of input-output variables
        for input in inputs:
            # prepare new dst_flow for gate-input port
            if gate.clock_domain:
                tmp_flow = ([], set([gate.clock_domain]))
            else:
                tmp_flow = ([], set())
                debug4('%s.%s' % (gate.name, input.name))
                iterable_dst.append((gate, input))
            
            for output in outputs:
                tmp_flow[0].append(output)
                # add new dependency to src_flow
                if output in src_flow:
                    src_flow[output][0].add(input)
                else:
                    src_flow[output] = (set([input]), set())
                # add clock event setting signal output
                if gate.clock_domain:
                    src_flow[output][1].add(gate.clock_domain)
            dst_flow[(gate, input)] = tmp_flow

            # update the set of gates to which the signal is connected
            if input in signal_to_gates:
                signal_to_gates[input].add(gate)
            else:
                signal_to_gates[input] = set([gate])

    # substitute signal in a list of dst_flow with gate-input instead
    for wire, flow in dst_flow.items():
        signals = flow[0][:]
        for signal in signals:
            flow[0].remove(signal)
            if signal in signal_to_gates:
                for gate in signal_to_gates[signal]:
                    flow[0].append((gate, signal))

    debug4('\nSignals over which calculate source transitive closure')
    # look for signals sets by blank clock domain
    for signal, flow in src_flow.items():
        if not flow[1]: # if the set of clocks is empty
            iterable_src.append(signal)
            debug4(signal.name)

    # stick flows to model
    model.src_flow = src_flow
    model.dst_flow = dst_flow

    # for optimized transitive clocure calculation
    model.iterable_src = iterable_src
    model.iterable_dst = iterable_dst

def flow_transitive_closure(flow, iterable):
    'makes model.???_flow be transitive closure wrt. clock set'
    # model.???_flow: Var1/Wire1->([Var2/Wire2],set(clock))
    last_change = None
    for signal in iterable:
        # one step of trans. clos. calculation
        vars, clockset = flow[signal]
        old_len = len(clockset)
        for var in vars:
            # !!! FEATURE REQUEST: if user wants to extend also the input ports
            # of the whole component, var is not in model.src_flow and it must
            # have unique clock domain, e.g.,:
            # (Variable(uniqname(var.name+'_clock')), 'rising')
            # ! update also the code below (copied+pasted code)
            if var in flow:
                clockset |= flow[var][1]
        if old_len != len(clockset):
            last_change = signal

    iter = 0
    while last_change:
        iter += 1
        for signal in iterable:
            if last_change == signal:
                last_change = None
                break
            # one step of trans. clos. calculation
            vars, clockset = flow[signal]
            old_len = len(clockset)
            for var in vars:
                if var in flow:
                    clockset |= flow[var][1]
            if old_len != len(clockset):
                last_change = signal

    return iter

def wire2str(wire):
    'debug purpose'
    return '%s.%s' % (wire[0].name, wire[1].name)

def print_dst_flow(model):
    'prints table of destination signal flows'
    def critical_mark(wire):
        'returns character symbol for critical attribute of a port'
        try:
            return '*' if wire in model.critical_ports else ' '
        except AttributeError:
            return ' '

    debug4('Destination input ports flow (* = critical input port)')
    #debug4('%40s | %-80s | %s' % ('input port', 'dependency', 'edges'))
    debug4('  %15s | %-40s | %s' % \
        ('input port', 'successors', 'dest. CDs'))
    debug4('-'*76)
    items = model.dst_flow.items()
    items.sort(lambda x, y: 1 if x[0][0].name > y[0][0].name else
        1 if x[0][0].name == y[0][0].name and x[0][1].name > y[0][1].name else
        -1)
    for wire, flow in items:
        #debug4('%40s | %-80s | %s' % (
        debug4('%s %15s | %-40s | %s' % (
            critical_mark(wire),
            wire2str(wire),
            ', '.join([wire2str(w) for w in flow[0]]),
            ', '.join([clk2str(c) for c in flow[1]])))
    

def print_src_flow(src_flow):
    'prints table of source signal flows'
    debug4('Source signals flow')
    debug4('%15s | %-40s | %s' % \
        ('signal', 'predecessors', 'source CDs'))
    debug4('-'*76)
    items = src_flow.items()
    items.sort(lambda x, y: 1 if x[0].name > y[0].name else -1)
    for signal, flow in items:
        debug4('%15s | %-40s | %s' % (
            signal.name,
            ', '.join([v.name for v in flow[0]]),
            ', '.join(clk2str(c) for c in flow[1])))

def is_synchronous(gate):
    'returns if an instance is of synchronous module'
    return len(gate.module.clocks) == 1

def make_cdc_length_table(model):
    'makes table mode.cdc_lengths: clock -> int'
    def get_max_length(wire):
        if wire[0] in clock_domain:
            return int(wire in model.critical_ports)
        succs, d_clocks = model.dst_flow[wire]
        if clock not in d_clocks:
            return 0
        max = 0
        for succ in succs:
            length = get_max_length(succ)
            if length > max:
                max = length
        return int(wire in model.critical_ports) + max

    model.cdc_table = {}
    for clock, clock_domain in model.clock_domains.items():
        max = 0
        for wire, (succs, clockset) in model.dst_flow.items():
            if wire[0] in clock_domain:
                for succ in succs:
                    length = get_max_length(succ)
                    if length > max:
                        max = length
        model.cdc_table[clock] = max
    return model.cdc_table

def print_cdc_lengths(cdc_lengths):
    '''prints clock period wrt. the longest critical path from one domain,
    intersecting other, ending the original clock domain'''
    debug2('Clock constraint')
    debug2('%20s | %s' % ('Clock domain', 'minimal clock period'))
    debug2('-'*76)
    for clock, length in cdc_lengths.items():
        debug2('%20s | %d step(s)' % (clk2str(clock), length))

def is_critical_port(model, wire):
    'returns True if wire is critical input port'
    # source clock domains differs from destination clock domains
    return (wire[1] in model.src_flow and wire in model.dst_flow) and \
           (model.src_flow[wire[1]][1] != model.dst_flow[wire][1])

def get_critical_ports(model):
    'returns and set model.critical_ports'
    critical = []
    for wire in model.dst_flow:
        if is_critical_port(model, wire):
            critical.append(wire)
    critical.sort(lambda x, y: 1 if x[0].name > y[0].name else
        1 if x[0].name == y[0].name and x[1].name > y[1].name else
        -1) 
    model.critical_ports = critical
    return critical

