#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: method_destabil.py 666 2009-09-14 16:01:03Z smrcka $
# Due to analysis of clock domain crossing performed on one module only, the
# algorithm implemented in this python module does not support the application
# of the method of extending every input port of a critical signal path
# on a hardware design containing more than one modules sensitive on more than
# 2 clock signals.
'extending all critical signal paths (with destabilizers)'

import common
import model as m
from math import log, ceil # calculation of clock counter size in bits

class DestabilizerModule(object):
    'a string reprezenting module definition'
    modules = {}
    def __init__(self, ninputs, length):
        self.ninputs = ninputs
        self.name = 'destabilizer_n%d_l%d' % (ninputs, length)
        self.length = length
        DestabilizerModule.modules[(ninputs, length)] = self

    def __str__(self):
        inputs = ['i%d' % i for i in range(self.ninputs)]
        return '''module %(modname)s(%(inputs)s, alpha, omega) {
    %(inpdecl)s
    input alpha: boolean;
    output omega: boolean;
    cnt: 0..%(L)d;

    init(cnt) := 0;                      -- D state
    next(cnt) := case {
        %(inpcond)s
                              : 1;       -- to X state
        (cnt = %(L)d)             : 0;       -- to D state
        (cnt > 0) & (cnt < %(L)d) : cnt + 1; -- remain in X
        1                     : 0;       -- remain in D
    };

    omega := case {
        (cnt = 0) : alpha;
        (cnt > 0) : {0, 1};
    };
}
''' % {
    'modname': self.name,
    'inputs': ', '.join(inputs),
    'inpdecl': '\n    '.join(['input %s: boolean;' % i for i in inputs]),
    'inpcond': '|\n        '.join(['(%s ^ next(%s))' % (i, i) for i in inputs]),
    'L': self.length
}

class DestabilizerInstStmt(object):
    'a class representing the SMV module instance statement of a destabilizer'
    def __init__(self, alpha, omega, inputs, length):
        self.name = common.uniqname('destabilizer')
        self.alpha = alpha
        self.omega = omega
        self.inputs = inputs
        self.module = DestabilizerModule(len(inputs), length)

    def __str__(self):
        return '%s: %s(%s, %s, %s);' % ( \
            self.name,
            self.module.name,
            ', '.join([str(i) for i in self.inputs]),
            self.alpha,
            self.omega)

def extend(model):
    "apply the method of extending all critical input ports"
    # flatten every "renaming module", i.e., every module containing the 
    # simplest assignment statements only 
    common.flatten_renaming_modules(model)

    # let only one module being sensed by more than one clock signals
    common.flatten_all_asyn_modules(model)

    # transform every assignment statement representing combinational logic to
    # structural description
    common.comb_logic_to_structural(model)
    common.debug2('Module clocks')
    common.print_found_clocks(model, 3)

    # find a module to extend
    model.asyn_module = common.find_module_to_extend(model)
    if not model.asyn_module:
        return None

    # find all instantiated modules and set them to appropriate clock domain
    common.debug1('Partitioning submodules of "%s" to clock domains' % \
        model.asyn_module.name)
    common.partition_clock_domain(model)
    common.print_clock_domains(model)

    # get the list of all module instances
    model.gates = common.get_modinst_list(model)

    # make soruce and destination direct data flow
    common.make_src_dst_flow(model)

    # make transitive closure of both flow sets
    common.debug1('Searching for critical signal paths')
    iter = common.flow_transitive_closure(model.src_flow, model.iterable_src)
    common.print_src_flow(model.src_flow)
    common.debug4('Source flow iterations: %d' % iter)
    iter = common.flow_transitive_closure(model.dst_flow, model.iterable_dst)
    common.debug4('Destination flow iterations: %d' % iter)
    common.get_critical_ports(model)
    common.print_dst_flow(model)

    # search for critical signal paths
    search_all_paths(model)
    print_critical_paths(model)

    # extend all paths
    model.new_bits = 0
    extend_all_paths(model)
    common.debug1('Extended %d paths' % len(model.critical_paths))

    # make table of length of clock domain crossing signal paths
    cdc_lengths = make_cdc_lengths(model)
    common.print_cdc_lengths(cdc_lengths)

    # constrain clock signals
    common.debug1('Constraining clock signals behavior')
    constrain_clocks(model)

    common.debug1('Introduced new %d bits of state variables' % model.new_bits)

    return model

def search_all_paths(model):
    'extends all critical signal path outputs'

    def get_inputs(signal, clock_domain):
        '''goes backward to other clock domain, counts the critical gates and
        returns all input signals'''
        if signal in model.src_flow:
            src_signals, src_clocks = model.src_flow[signal]
        else:
            # TODO: check here, if a signal is component input or has a
            # constant value. If it is an input, it can also be a critical port
            # to the whole component - so return ([signal], 1) instead.
            return ([], 0)
        if src_clocks == clock_domain:
            return ([], 0)
        elif len(src_clocks) == 1: # signal is set synchronously
            return ([signal], 1)
        else:
            inputs = []
            max_length = 0
            for src_signal in src_signals:
                inp, length = get_inputs(src_signal, clock_domain)
                inputs += inp
                if length > max_length:
                    max_length = length
            return (inputs, max_length + 1)
            
    paths = {}
    for gate, signal in model.critical_ports:
        # if critical input port is on a synchronous gate
        if gate.clock_domain:
            # go backward to other clock domain
            paths[(gate,signal)] = get_inputs(signal, set([gate.clock_domain]))
    model.critical_paths = paths

def print_critical_paths(model):
    'prints a table of critical paths ant their inputs'
    common.debug2('Critical paths')
    common.debug2('%15s | %-40s | %s' % ('path output', 'inputs', 'length'))
    common.debug2('-'*76)
    for wire, (inputs, length) in model.critical_paths.items():
        common.debug2('%15s | %-40s | %d' % (common.wire2str(wire),
            ', '.join([v.name for v in inputs]),
            length))

def make_cdc_lengths(model):
    '''makes table model.cdc_table: clock -> int being the maximal possible on
    clock period caused by a destabilizer'''
    cdc_lengths = dict([(clock, 0) for clock in model.clock_domains])
    for (gate, signal), (inputs, length) in model.critical_paths.items():
        dst_clock = gate.clock_domain
        if length > cdc_lengths[dst_clock]:
            cdc_lengths[dst_clock] = length
    model.cdc_table = cdc_lengths
    return cdc_lengths

def constrain_clocks(model):
    'create a counter for every constraint clock signal, constrain the signal'
    def constraint(clk, edge, length, stmt, clk_en_stmt, clk_val):
        'contrain one clock wrt. minimal period steps'
        # new counter
        cnt_type = m.IntType(0, length)
        cnt = m.Variable(common.uniqname('%s_counter' % clk.name), cnt_type)
        cnt_decl = m.VariableDeclStmt(cnt)
        # counter behaviour
        cnt_init = m.AssignmentStmt(m.InitOp(cnt),
            m.Constant(length, m.INT),
            False)
        if edge == 'rising':
            edge_expr = m.Operator('&',
                [m.UnaryOp('~', clk, False), m.NextOp(clk)], False)
        else:
            edge_expr = m.Operator('&',
                [clk, m.UnaryOp('~', m.NextOp(clk), False)], False)
        cnt_next = m.AssignmentStmt(m.NextOp(cnt),
            m.CaseBlock([
                (m.Operator('<', [cnt, m.Constant(length, m.INT)], False),
                 m.Operator('+', [cnt, m.Constant(1, m.INT)], False)),
                (edge_expr, m.Constant(0, m.INT)),
                (m.Constant(1, m.INT), cnt),
            ], False),
            False)
        # clk constraint
        clk_active_edge = clk if edge == 'rising' else m.UnaryOp('~', clk,
        False)
#        clk_enable = m.Operator('|', [
#            clk_active_edge, m.Operator('=', [
#                cnt,
#                m.Constant(length, m.INT)], False)], False)
        clk_enable = m.Operator('=', [cnt, m.Constant(length, m.INT)], False)
        if clk_en_stmt:
            clk_en_stmt.src = clk_enable
            clk_enable = clk_en_stmt.dst
        if not clk_val:
            clk_val = m.Constant(m.SetType([0,1]), m.INT)
        stmt.src = m.CaseBlock([
            (clk_enable, clk_val),
            (m.Constant(1, m.INT), m.Constant(0, m.INT))], False)
        stmt.dst = m.NextOp(stmt.dst)
        return [cnt_decl, [cnt_init, cnt_next]]

    counters = []
    constraints = []
    for (clk, edge), length in model.cdc_table.items():
        if not length: continue
        for stmt in model.asyn_module.stmts:
            if isinstance(stmt, m.AssignmentStmt) and stmt.dst == clk:
                clk_en = None
                clk_val = None
                ce_stmt = None
                if not isinstance(stmt.src, m.Constant):
                    # clock value is conditioned in original model
                    deps = common.get_vars(stmt.src)
                    if len(deps) >= 1:
                        if deps[0].name in common.OPTS['clk_enable']:
                            clk_en = deps[0]
                        elif common.OPTS['auto_clk']:
                            clk_en = deps[0]
                            common.debug1(
                         "warning: assuming `%s' as enable signal for `%s'" % \
                                (deps[0].name, clk.name))
                        else:
                            raise NotImplementedError( \
                        "unable to constrain clock `%s'\n"
                        "depends on: %s\n"
                        "suggestion: change model or try -c or -a" % \
                              (clk.name, ', '.join([v.name for v in deps])))
                    if len(deps) == 2:
                        if deps[1].name in common.OPTS['clk_val']:
                            clk_val = deps[1]
                        elif common.OPTS['auto_clk']:
                            clk_val = deps[1]
                            common.debug(0,
                          "warning: assuming `%s' as value signal for `%s'" % \
                                (deps[1].name, clk.name))
                        else:
                            raise NotImplementedError( \
                                "unable to constrain clock `%s'\n"
                                "don't know what to do with `%s'\n"
                                "suggestion: change model or try -e or -a" % \
                                (clk.name, deps[1].name))
                    else:
                        raise NotImplementedError( \
                        "unable to constrain clock `%s'\n"
                        "depends on: %s\n"
                        "suggestion: change model (even -a won't help)" % \
                          (clk.name, ', '.join([v.name for v in deps])))
                   
                    if clk_en:
                        # find clock enable assignment statement
                        for stmt1 in model.asyn_module.stmts:
                            if isinstance(stmt1, m.AssignmentStmt) and \
                              stmt1.dst == clk_en:
                                ce_stmt = stmt1
                                break
                # constrain clock assignment and clock enable assignment
                counter, stmts = constraint(clk,
                    edge,
                    length,
                    stmt,
                    ce_stmt,
                    clk_val)
                counters.append(counter)
                constraints += stmts
                model.new_bits += ceil(log(length+1)/log(2))
    for counter in counters:
        model.asyn_module.stmts.insert(0, counter)
    model.asyn_module.stmts += constraints

def extend_path(model, gate, paramname, signal, inputs, length):
    'extend the end of critical signal path with a new destabilizer'
    if not gate.param_map[paramname].type.compatible(m.BOOLEAN):
        raise NotImplementedError( \
          'unable to extend non-boolean type input port `%s` of module `%s`'  \
            % (paramname, modinst.module.name))
    # create omega signal
    omega = m.Variable(common.uniqname('omega_%s' % signal.name), m.BOOLEAN)
    omega_decl = m.VariableDeclStmt(omega)
    model.asyn_module.stmts.insert(0, omega_decl)
    # connect it to input port
    gate.param_map[paramname] = omega
    # create a destabilizer
    destabilizer = DestabilizerInstStmt(signal, omega, inputs, length)
    model.asyn_module.stmts.append(destabilizer)

def extend_all_paths(model):
    'extends every critical signal path with destabilizer'
    for (gate, signal), (inputs, length) in model.critical_paths.items():
        # find a name of an input port
        for paramname, connected in gate.param_map.items():
            if signal in common.get_vars(connected):
                extend_path(model, gate, paramname, signal, inputs, length)
                model.new_bits += ceil(log(length+1)/log(2))
    for destabilizer in DestabilizerModule.modules.values():
        model.stmts.append(destabilizer)

#------------------ TO BE REMOVED ------------------------------------------

def extend_gate_input(model, gate, paramname, signal):
    '''extends the signal path ending in `modinst`-`input` wrt clock domain
    of `clock` with delayed input modules'''
    if not gate.param_map[paramname].type.compatible(m.BOOLEAN):
        raise NotImplementedError( \
          'unable to extend non-boolean type input port `%s` of module `%s`' \
            % (paramname, modinst.module.name))
    # create new signal connecting input-delay with gate
    d_var = m.Variable(common.uniqname('delay_%s' % signal.name), m.BOOLEAN)
    d_decl = m.VariableDeclStmt(d_var)
    model.asyn_module.stmts.insert(0, d_decl)
    # connect it to input port
    gate.param_map[paramname] = d_var
    # create input-delay
    di_inst = DelayedInputModInst(signal, d_var)
    model.asyn_module.stmts.append(di_inst)

def extend_critical_input_ports(model):
    '''extends every connection of model.critical_ports to a gate with a new
    instance of input-delay'''
    model.critical_ports_number = 0
    for gate, signal in model.critical_ports:
        # find name of an input port
        for paramname, connected in gate.param_map.items():
            if signal in common.get_vars(connected):
                extend_gate_input(model, gate, paramname, signal)
                model.critical_ports_number += 1
    if model.critical_ports:
        model.stmts.append(DelayedInputModule())

