#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: method_dinput.py 666 2009-09-14 16:01:03Z smrcka $
# Due to analysis of clock domain crossing performed on one module only, the
# algorithm implemented in this python module does not support the application
# of the method of extending every input port of a critical signal path
# on a hardware design containing more than one modules sensitive on more than
# 2 clock signals.
'extending all critical input ports (with delayed inputs)'

import common
import model as m
from math import log, ceil


class DelayedInputModule(object):
    'delayed input module SMV implementation'
    name = 'delayed_input'
    def __str__(self):
        return '''module %s(signal, delayed)
{
    input signal : boolean;
    output delayed : boolean;

    init(delayed) := signal;
    if (signal = next(signal))
        next(delayed) := signal;
    else
        next(delayed) := {0, 1};
}''' % self.name

class DelayedInputModInst(object):
    'delayed input module instance statement'
    def __init__(self, input, output):
        self.name = common.uniqname('delayed_input')
        self.input = input
        self.output = output
    
    def __str__(self):
        return '%s : %s(%s, %s);' % (self.name, DelayedInputModule.name, \
            self.input, self.output)

def extend(model):
    "apply the method of extending all critical input ports"
    # flatten every "renaming module", i.e., every module containing the 
    # simplest assignment statements only 
    common.flatten_renaming_modules(model)

    # let only one module being sensed by more than one clock signals
    common.flatten_all_asyn_modules(model)

    # transform every assignment statement representing combinational logic to
    # structural description
    common.comb_logic_to_structural(model)
    common.debug2('Module clocks')
    common.print_found_clocks(model, 3)

    # find a module to extend
    model.asyn_module = common.find_module_to_extend(model)
    if not model.asyn_module:
        return None

    # find all instantiated modules and set them to appropriate clock domain
    common.debug1('Partitioning submodules of "%s" to clock domains' % \
        model.asyn_module.name)
    common.partition_clock_domain(model)
    common.print_clock_domains(model)

    # get the list of all module instances
    model.gates = common.get_modinst_list(model)

    # make soruce and destination direct data flow
    common.make_src_dst_flow(model)

    # make transitive closure of both flow sets
    common.debug1('Searching for critical signal paths')
    iter = common.flow_transitive_closure(model.src_flow, model.iterable_src)
    common.print_src_flow(model.src_flow)
    common.debug4('Source flow iterations: %d' % iter)
    iter = common.flow_transitive_closure(model.dst_flow, model.iterable_dst)
    common.debug4('Destination flow iterations: %d' % iter)
    common.get_critical_ports(model)
    common.print_dst_flow(model)

    # make table of length of clock domain crossing signal paths
    cdc_lengths = common.make_cdc_length_table(model)
    common.print_cdc_lengths(cdc_lengths)

    # extend all critical input ports
    extend_critical_input_ports(model)
    common.debug1('Extended critical input ports: %d' % \
        model.critical_ports_number)
    model.new_bits = model.critical_ports_number

    # constrain clock signals
    common.debug1('Constraining clock signals behavior')
    constrain_clocks(model)

    common.debug1('Introduced new %d bits of state variables' % model.new_bits)

    return model

def extend_gate_input(model, gate, paramname, signal):
    '''extends the signal path ending in `modinst`-`input` wrt clock domain
    of `clock` with delayed input modules'''
    if not gate.param_map[paramname].type.compatible(m.BOOLEAN):
        raise NotImplementedError( \
          'unable to extend non-boolean type input port `%s` of module `%s`' \
            % (paramname, modinst.module.name))
    # create new signal connecting input-delay with gate
    d_var = m.Variable(common.uniqname('delay_%s' % signal.name), m.BOOLEAN)
    d_decl = m.VariableDeclStmt(d_var)
    model.asyn_module.stmts.insert(0, d_decl)
    # connect it to input port
    gate.param_map[paramname] = d_var
    # create input-delay
    di_inst = DelayedInputModInst(signal, d_var)
    model.asyn_module.stmts.append(di_inst)

def extend_critical_input_ports(model):
    '''extends every connection of model.critical_ports to a gate with a new
    instance of input-delay'''
    model.critical_ports_number = 0
    for gate, signal in model.critical_ports:
        # find name of an input port
        for paramname, connected in gate.param_map.items():
            if signal in common.get_vars(connected):
                extend_gate_input(model, gate, paramname, signal)
                model.critical_ports_number += 1
    if model.critical_ports:
        model.stmts.append(DelayedInputModule())

def constrain_clocks(model):
    'create a counter for every constraint clock signal, constrain the signal'
    def constraint(clk, edge, length, stmt, clk_en_stmt, clk_val):
        'contrain one clock wrt. minimal period steps'
        # new counter
        cnt_type = m.IntType(0, length)
        cnt = m.Variable(common.uniqname('%s_counter' % clk.name), cnt_type)
        cnt_decl = m.VariableDeclStmt(cnt)
        # counter behaviour
        cnt_init = m.AssignmentStmt(m.InitOp(cnt),
            m.Constant(length, m.INT),
            False)
        if edge == 'rising':
            edge_expr = m.Operator('&',
                [m.UnaryOp('~', clk, False), m.NextOp(clk)], False)
        else:
            edge_expr = m.Operator('&',
                [clk, m.UnaryOp('~', m.NextOp(clk), False)], False)
        cnt_next = m.AssignmentStmt(m.NextOp(cnt),
            m.CaseBlock([
                (m.Operator('<', [cnt, m.Constant(length, m.INT)], False),
                 m.Operator('+', [cnt, m.Constant(1, m.INT)], False)),
                (edge_expr, m.Constant(0, m.INT)),
                (m.Constant(1, m.INT), cnt),
            ], False),
            False)
        # clk constraint
        clk_active_edge = clk if edge == 'rising' else m.UnaryOp('~', clk,
        False)
#        clk_enable = m.Operator('|', [
#            clk_active_edge, m.Operator('=', [
#                cnt,
#                m.Constant(length, m.INT)], False)], False)
        clk_enable = m.Operator('=', [cnt, m.Constant(length, m.INT)], False)
        if clk_en_stmt:
            clk_en_stmt.src = clk_enable
            clk_enable = clk_en_stmt.dst
        if not clk_val:
            clk_val = m.Constant(m.SetType([0,1]), m.INT)
        stmt.src = m.CaseBlock([
            (clk_enable, clk_val),
            (m.Constant(1, m.INT), m.Constant(0, m.INT))], False)
        stmt.dst = m.NextOp(stmt.dst)
        return [cnt_decl, [cnt_init, cnt_next]]

    counters = []
    constraints = []
    for (clk, edge), length in model.cdc_table.items():
        if not length: continue
        for stmt in model.asyn_module.stmts:
            if isinstance(stmt, m.AssignmentStmt) and stmt.dst == clk:
                clk_en = None
                clk_val = None
                ce_stmt = None
                if not isinstance(stmt.src, m.Constant):
                    # clock value is conditioned in original model
                    deps = common.get_vars(stmt.src)
                    if len(deps) >= 1:
                        if deps[0].name in common.OPTS['clk_enable']:
                            clk_en = deps[0]
                        elif common.OPTS['auto_clk']:
                            clk_en = deps[0]
                            common.debug1(
                         "warning: assuming `%s' as enable signal for `%s'" % \
                                (deps[0].name, clk.name))
                        else:
                            raise NotImplementedError( \
                        "unable to constrain clock `%s'\n"
                        "depends on: %s\n"
                        "suggestion: change model or try -c or -a" % \
                              (clk.name, ', '.join([v.name for v in deps])))
                    if len(deps) == 2:
                        if deps[1].name in common.OPTS['clk_val']:
                            clk_val = deps[1]
                        elif common.OPTS['auto_clk']:
                            clk_val = deps[1]
                            common.debug(0,
                          "warning: assuming `%s' as value signal for `%s'" % \
                                (deps[1].name, clk.name))
                        else:
                            raise NotImplementedError( \
                                "unable to constrain clock `%s'\n"
                                "don't know what to do with `%s'\n"
                                "suggestion: change model or try -e or -a" % \
                                (clk.name, deps[1].name))
                    else:
                        raise NotImplementedError( \
                        "unable to constrain clock `%s'\n"
                        "depends on: %s\n"
                        "suggestion: change model (even -a won't help)" % \
                          (clk.name, ', '.join([v.name for v in deps])))
                   
                    if clk_en:
                        # find clock enable assignment statement
                        for stmt1 in model.asyn_module.stmts:
                            if isinstance(stmt1, m.AssignmentStmt) and \
                              stmt1.dst == clk_en:
                                ce_stmt = stmt1
                                break
                # constrain clock assignment and clock enable assignment
                counter, stmts = constraint(clk,
                    edge,
                    length,
                    stmt,
                    ce_stmt,
                    clk_val)
                counters.append(counter)
                constraints += stmts
                model.new_bits += ceil(log(length+1)/log(2))
    for counter in counters:
        model.asyn_module.stmts.insert(0, counter)
    model.asyn_module.stmts += constraints
