#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: method_osd.py 664 2009-09-11 09:45:10Z smrcka $
# Due to analysis of clock domain crossing performed on one module only, the
# algorithm implemented in this python module does not support the application
# of the method of extension by one-step destabilizer on a hardware design
# containing more than one module sensitive on more than 2 clock signals.
'extending all critical signal paths (with one-step destabilizers)'

import common
import model as m

class OneStepDestabilizerModule(object):
    'A string class representing module definition'
    modules = {}
    def __init__(self, ninputs, edge):
        self.ninputs = ninputs
        self.edge = edge
        self.name = 'onestep_destabilizer_%d_%s' % (ninputs, edge)
        OneStepDestabilizerModule.modules[(ninputs, edge)] = self

    def __str__(self):
        inputs = ['i_%d' % i for i in range(self.ninputs)]
        return '''module %s(%s, clk, alpha, omega) {
    %s
    input clk : boolean;
    input alpha : boolean;
    output omega : boolean;

    init(omega) := alpha;

    next(omega) := case {
--    omega := case {
        (%s) : {0,1};
        1 : alpha;
    };
}
''' % ( \
    self.name, 
    ', '.join(inputs),
    '\n    '.join(['input %s : boolean;' % i for i in inputs]),
    ' | '.join(['(%s ^ next(%s))' % (i, i) for i in inputs]),
#    '~clk & next(clk)' if self.edge == 'rising' else \
#        'clk & ~next(clk)',
#--        (%s) &
#--        (%s) : {0,1};
#    '~next(clk) & next(next(clk))' if self.edge == 'rising' else \
#        'next(clk) & ~next(next(clk))',
)


class OneStepDestabilizerInstStmt(object):
    '''a class representing the SMV module instance statement of one-step
    destabilizer'''
    def __init__(self, clock, alpha, omega, inputs):
        self.name = common.uniqname('destabil')
        self.clock = clock  # (Variable, str_edge_type)
        self.alpha = alpha  # Expr
        self.omega = omega  # Variable
        self.inputs = inputs # list of Variables
        self.module = OneStepDestabilizerModule(len(inputs), clock[1])

    def __str__(self):
        return '%s : onestep_destabilizer_%d_%s(%s, %s, %s, %s);' % ( \
            self.name,
            len(self.inputs),
            self.clock[1],
            ', '.join([str(i) for i in self.inputs]),
            self.clock[0],
            self.alpha,
            self.omega,
        )
        

def create_osd(clock, alpha, inputs):
    '''create one-step destabilizer module and return its instance statement
    dest_clk is destination clock structure (Variable, edge) - c_2
    alpha is an expression
    inputs is a list of monitored inputs (i_1,...,i_n)
    Returns (VariableDecl-omega, ModuleInstStmt-destabil)'''
    omega = m.Variable(common.uniqname('omega'), m.BOOLEAN)
    decl_omega = m.VariableDeclStmt(omega)
    modinst = OneStepDestabilizerInstStmt(clock, alpha, omega, inputs)
    return (decl_omega, modinst)

def extend(model):
    "apply the method of extending all critical input ports"
    # flatten every "renaming module", i.e., every module containing the 
    # simplest assignment statements only 
    common.flatten_renaming_modules(model)

    # let only one module being sensed by more than one clock signals
    common.flatten_all_asyn_modules(model)

    #!!!
    common.comb_logic_to_structural(model)

    # find a module to extend
    model.asyn_module = common.find_module_to_extend(model)
    if not model.asyn_module:
        return None

    # find all instantiated modules and set them to appropriate clock domain
    common.debug1('Partitioning submodules of "%s" to clock domains' % \
        model.asyn_module.name)
    common.partition_clock_domain(model)
    common.print_clock_domains(model)

    common.debug1('Marking every variable wrt. clock sensitivity')
    # create variable sensitivity map
    common.create_sens_map(model)
    # mark sensitivity of every output port of a synchronous domain and for
    # all alias variable of that ports
    common.add_output_port_sens(model)
    # union all the clock sensitivity of array items and make the whole array
    # being sensitive on that union
    common.add_array_sens(model)

    # make a map model.trans_dep Variable->[Variable] to get the info which
    # variable is set by which variable - needed when connecting monitored
    # input of a destabilizer
    common.create_transparent_deps(model)

    # resolve dependencies of clock sensitivity among all statements
    # make a transitive closure of model.sens
    c_sens = common.copy_sens_map(model.sens)
    common.set_dep_sens(model)
    iteration = 1
    while c_sens != model.sens:
        common.debug2('Resolving dependencies - iteration %d' % iteration)
        iteration += 1
        c_sens = common.copy_sens_map(model.sens)
        common.set_dep_sens(model)
    common.print_sens_map(model)

    # extend every input port from uncompatible clock domain
    extend_module_instances(model)

    # finally add the one-step destabilizer module definitions
    for module in OneStepDestabilizerModule.modules.values():
        model.stmts.append(module)

    return model

def extend_iport_with_destabil(model, modinst, portname, clock):
    'extends an input port with a one-step destabilizer'
    alpha = modinst.param_map[portname]
    # get inputs of critical signal path
    inputs = set()
    for var in common.get_vars(alpha):
#        common.print_trans_dep(model, var)

        monitored_inputs = [v for v in model.trans_dep[var]
            if model.sens[v] != set([clock])]
        inputs = inputs.union(set(monitored_inputs))
        #inputs = inputs.union(set(model.trans_dep[var]))
    decl_omega, destabil_modinst = create_osd(clock, alpha, list(inputs))
    modinst.param_map[portname] = decl_omega.variable

    model.asyn_module.stmts.insert(0, decl_omega)
    model.asyn_module.stmts.append(destabil_modinst)

def extend_module_instance(model, modinst, clock):
    'extends all uncompatible input ports'
    for portname, portexpr in modinst.param_map.items():
        if modinst.module.parameters[portname].direction == 'input':
            for var in common.get_vars(portexpr):
                sens = model.sens[var]
                if sens and sens != set([clock]):
                    common.debug1('%-30s | %-12s | %-20s' % \
                        (modinst.name, modinst.module.name+'.'+portname, var))
                    extend_iport_with_destabil(model, modinst, \
                        portname, clock)
                    break   # further search at this port is not needed

def extend_module_instances(model):
    'extend every synchronized module instance'
    # for every module in all clock domains, extend its input ports with
    # destabilizers if it is on the output of the critical signal path
    # resolve dependencies of clock sensitivity among all statements
    common.debug1('Extending ends of critical signal paths')
    common.debug1('%-30s | %-12s | %-20s' % \
        ('instance', 'module.port', 'signal'))
    common.debug1('-'*75)
    for clock, mlist in model.clock_domains.items():
        for modinst in mlist:
            extend_module_instance(model, modinst, clock)
