#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id: model.py 651 2009-09-01 17:04:34Z smrcka $
"""model.py - model structure and model modification methods"""

import re
#from common import log

class SMVModuleError(Exception):
    'an exception bound to SMV language errors'
    pass

class Model(object):
    'model structure'
    def __init__(self):
        self.stmts = []     # list of statements
        self.modules = []   # list of Module, current: modules[-1]
        self.main_module = None # the main module
        self.clock_domains = {} # clock domains: clock->list of ModuleInstStmt

    def __str__(self):
        return stmt_list2str(self.stmts, '')
        
    def find_id(self, name):
        'find an identifier in current module or global scope'
        # search current module
        if len(self.modules):
            obj = self.modules[-1].find_id(name)
            if obj:
                return obj
        # search global types
        types = [s for s in self.stmts if isinstance(s, TypeDefStmt)]
        for stmt in types:
            if stmt.name == name:
                return stmt
        # search global constants
        for stmt in types:
            if stmt.type == SetType:
                for const in stmt.type.values:
                    if const == name:
                        return Constant(const, INT)

    def set_main_module(self):
        'find and set the main module'
        try:
            self.main_module = [module \
                for module in self.modules if module.name == 'main'][0]
        except IndexError:
            raise SMVModuleError("unable to find 'main' module")
        return self.main_module

    def set_used_modules(self):
        'find and set modules really used from main_module - set the .modules'
        def get_used_modules(module):
            'find instantiated modules recursively'
            mods = []
            for stmt in module.stmts:
                if isinstance(stmt, ModuleInstStmt):
                    mods.append(stmt.module)
                    mods += get_used_modules(stmt.module)
            return mods

        self.modules = [self.main_module]
        # uniquely add
        for mod in get_used_modules(self.main_module):
            if mod not in self.modules:
                self.modules.append(mod)

        # update .stmts
        stmts = []
        for stmt in self.stmts:
            if isinstance(stmt, Module):
                if stmt in self.modules:
                    stmts.append(stmt)
            else:
                stmts.append(stmt)
        self.stmts = stmts

class AbstractType(object):
    'class of abstact type of variables'
    def __init__(self, name=''):
        self.name = name

    def __str__(self):
        return id2smv(self.name)

    def __len__(self):
        raise NotImplementedError(str(type(self)))

    def compatible(self, utype):
        'returns true if a given type is compatible'
        raise NotImplementedError(str(type(self)))
    
    def mkinstance(self, name_prefix):
        'makes a variable instance of this type'
        return Variable(name_prefix, self, False)

class UndefinedType(AbstractType):
    'undefined type (undefined SMV keyword)'
    def __init__(self):
        AbstractType.__init__(self, 'undefined')

    def __len__(self):
        return 0

    def compatible(self, utype):
        'returns true cause it undefined is always compatible'
        return True

UNDEFINED = UndefinedType()

class IntType(AbstractType):
    'integer type of a variable'
    def __init__(self, lower, upper, name=''):
        AbstractType.__init__(self, name)
        self.lower = lower
        self.upper = upper

    def __str__(self):
        return id2smv(self.name) if self.name else \
            '%s..%s' % (self.lower, self.upper)

    def __len__(self):
        return -1

    def compatible(self, utype):
        'returns true if utype is compatible'
        return isinstance(utype, IntType)
    
INT = IntType(0, 0)

class BooleanType(IntType):
    'boolean type of a variable (singleton)'
    def __init__(self, name='boolean'):
        IntType.__init__(self, 0, 1, name)

    def __len__(self):
        return 1

    def __str__(self):
        return self.name

BOOLEAN = BooleanType()

class SetType(IntType):
    """a type defining a set of constants/identifiers. Argument values is
    the list of identifiers"""
    def __init__(self, values):
        IntType.__init__(self, 0, len(values)-1, '')
        self.values = values

    def __str__(self):
        return id2smv(self.name) if self.name else \
            '{%s}' % ', '.join([id2smv(str(i)) for i in self.values])

    def __len__(self):
        return -1

class ArrayType(AbstractType):
    'array type'
    def __init__(self, name, lower, upper, basetype):
        AbstractType.__init__(self, name)
        try:
            self.lower = int(lower)
            self.upper = int(upper)
        except TypeError:
            raise TypeError('array boundary not defined by integers')
        self.basetype = basetype
        self.items = []

    def __len__(self):
        return abs(self.upper - self.lower) + 1

    def __str__(self):
        return id2smv(self.name) if self.name else \
            'array %d..%d of %s' % (self.lower, self.upper, self.basetype)

    def compatible(self, utype):
        'returns true if compatible with utype'
        return isinstance(utype, ArrayType) and \
            self.basetype.compatible(utype.basetype)
#            len(self) == len(utype)

    def get_range(self):
        'returns an range object of array indexes'
        if self.lower <= self.upper:
            idx = range(self.lower, self.upper+1)
        else:
        #   we need var.data to be a list of indexes: [upper,upper-1,...,lower]
            idx = range(self.lower, self.upper-1, -1)
        return idx

    def mkinstance(self, name_prefix):
        'makes a variable instance of this type'
        return [self.basetype.mkinstance('%s[%s]' % (name_prefix, i)) \
            for i in self.get_range()]

BOOLARRAY = ArrayType('', 0, 0, BOOLEAN)

class Variable(object):
    'class representing SMV variable'
    def __init__(self, name, vtype, orig=True):
        self.name = name
        if not orig:
#            self.smv_name = id2smv(self.name)
#        else:
            self.smv_name = self.name
        self.type = vtype
        self.access = []
        if isinstance(self.type, ArrayType):
            self.data = self.type.mkinstance(id2smv(name))
        else:
            self.data = None

    def __str__(self):
        return self.smv_name

    def __setattr__(self, attr, value):
        self.__dict__[attr] = value
        if attr == 'name':
            self.__dict__['smv_name'] = id2smv(self.name)

    def rename_data_items(self, newname):
        'renames .name in every .data item'
        if isinstance(self.type, ArrayType):
            data_range = self.type.get_range()
            data_idx_range = range(len(data_range))
            for dri, dix in zip(data_range, data_idx_range):
                self.data[dix].name = '%s[%s]' % (newname, dri)
                self.data[dix].smv_name = '%s[%s]' % (id2smv(newname), dri)

    def copy(self):
        'returns a copy of self'
        nvar = Variable(self.name, self.type)
        return nvar

    def get_item(self, idx):
        'if the variable is of array type, return the idx-th item'
        if isinstance(self.type, ArrayType):
            if self.type.lower < self.type.upper:
                return self.data[idx-self.type.lower]
            else:
                return self.data[self.type.lower-idx]
        else:
            raise TypeError('unable to get an item from non-array type')

class ModuleParameter(Variable):
    'module parameter, i.e., name, type, direction'
    def __init__(self, name, ptype, direction):
        Variable.__init__(self, name, ptype)
        self.direction = direction # 'input'/'output'
        self.aliases = []   # the list of variables representing this port
                            # e.g.:
                            # input C: boolean;
                            # C_1: boolean;
                            # C_1 := C; -- C_1 is alias for C

class Module(object):
    'module declaration class'
    def __init__(self, name, paramnames):
        # paramnames - list of parameter names (str)
        # need to be explicitly declared with input/output within module body
        self.name = name
        self.paramnames = paramnames
        self.parameters = {} # str->ModuleParameter
        self.variables = {} # str->Variable
        self.stmts = [] # list of Statement

        self.aliases = [] # Variable x Variable relation of aliases wrt. simple
                          # assignment statement: var1 := var2;
        self.alias_classes = [] # equivalence classes of self.aliases

        self.clocks = []

    def __str__(self):
        return 'module %s(%s) {\n%s\n}' % \
            (id2smv(self.name),
            ', '.join([id2smv(p) for p in self.paramnames]),
             stmt_list2str(self.stmts))

    def find_id(self, name):
        'search for an object corresponding to the identifier'
        # search parameters
        try:
            return self.parameters[name]
        except KeyError:
            pass
        # search variables
        try:
            return self.variables[name]
        except KeyError:
            pass
        # search types and module instances
        for stmt in [s for s in self.stmts \
            if (isinstance(s, TypeDefStmt) or isinstance(s, ModuleInstStmt)) \
                and s.name == name]:
            return stmt
        # search set constants
        for stmt in [s for s in self.stmts \
            if isinstance(s, TypeDefStmt) and isinstance(s.type, SetType)]:
            for const in stmt.type.values:
                if const == name:
                    return Constant(const, INT)
        return None

    def mkinstance(self):
        '''makes an instance of the module, i.e., self variables and
        self module instances'''

        module = Module(self.name, self.paramnames)

        # make a new name->parameter map
        for name, param in self.parameters.items():
            module.parameters[name] = ModuleParameter(param.name, \
                param.type, param.direction)

        # make a new name->variable map
        for name, var in self.variables.items():
            module.variables[name] = Variable(var.name, var.type)

        # create the translation table for all local variables
        var_map = dict(module.variables.items() + module.parameters.items())
        # ...plus mapping of all array items
        for name, var in module.variables.items() + module.parameters.items():
            if isinstance(var.type, ArrayType):
                for item in var.data:
                    var_map[item.name] = item
        # replace statements wrt. their type, replace every occurence of the
        # local variable with newly created one (var_map translation table)
        module.stmts = [copy_stmt(s, var_map) for s in self.stmts]

        # copy alias table
        module.aliases = []
        for var1, var2 in self.aliases:
            module.aliases.append((var_map[var1.name], var_map[var2.name]))

        # copy clock sensitivity list
        module.clocks = []
        for clock, edge in self.clocks:
            try:
                module.clocks.append((var_map[clock], edge))
            except KeyError:
                pass

        return module

    def _add_alias_transitively(self, var1, var2):
        "add an alias tuple to .aliases preserving the transitive closure"
        # preserve reflexivity and symmetry
        new = [(var1, var1), (var2, var2), (var1, var2), (var2, var1)]
        # preserve transitivity
        # var1_2nd = { x | (var1, x) \in self.aliases}
        # var2_2nd = { y | (var2, y) \in self.aliases}
        var1_2nd = [varx for var1_, varx in self.aliases if var1_ == var1]
        var2_2nd = [varx for var2_, varx in self.aliases if var2_ == var2]
        # self.aliases' = self.aliases U 
        #                 new U
        #                 (var1_2nd x var2_2nd) U (var2_2nd x var1_2nd)
        for varx in var1_2nd:
            for vary in var2_2nd:
                new.append((varx, vary))
                new.append((vary, varx))
        # append new aliases
        for alias in new:
            if alias not in self.aliases:
                self.aliases.append(alias)

    def find_aliases(self):
        "fill up the self.aliases wrt. assignment statements"
        for stmt in self.stmts:
            if isinstance(stmt, AssignmentStmt) and \
              isinstance(stmt.dst, Variable) and \
              isinstance(stmt.src, Variable):
                self._add_alias_transitively(stmt.dst, stmt.src)

        self._make_alias_classes()

    def _make_alias_classes(self):
        '''makes .alias_classes from .aliases and sort them wrt. order of
        declaration statements'''

        # 1. fill-in the classes
        self.alias_classes = []
        closed = []
        for var1, var2 in self.aliases:
            tmp_class = []
            for varx in [x for v, x in self.aliases if v == var1]:
                if varx not in closed:
                    tmp_class.append(varx)
                    closed.append(varx)
            self.alias_classes.append(tmp_class)

        # 2. sort variables in classes
        # 2.1 get the order for sorting
        var_order = {}
        i = 0
        for stmt in self.stmts:
            if isinstance(stmt, VariableDeclStmt):
                var_order[stmt.variable] = i
                i += 1
                # also array items could be addressed directly
                # NOTE: two-and-more dimensional array fields unsupported
                if isinstance(stmt.variable.type, ArrayType):
                    for item in stmt.variable.data:
                        var_order[item] = i
                        i += 1
            elif isinstance(stmt, ModParamDeclStmt):
                var_order[stmt.parameter] = i
                i += 1
                # also array items could be addressed directly
                if isinstance(stmt.parameter.type, ArrayType):
                    for item in stmt.parameter.data:
                        var_order[item] = i
                        i += 1
        
        # 2.2 sort the item in alias classes
        for a_class in self.alias_classes:
            a_class.sort(lambda x, y: -1 if var_order[x] < var_order[y] else \
                1 if var_order[x] > var_order[y] else 0)

    def get_param_from_alias(self, var):
        """returns ModuleParameter within the alias domain of var
        returns None if ModuleParameter not found
        returns var if it is already a ModuleParameter"""
        if isinstance(var, ModuleParameter):
            return var
        for var1, var2 in self.aliases:
            if var1 == var and \
              var1 != var2 and \
              isinstance(var2, ModuleParameter):
                return var2
        return None

#=============================================================================
# Statement Classes
#
# Every statement object has .read_expr and .write_expr as the list of
# expressions which are, in the semantics of the statement, read or written
#=============================================================================

class Statement(object):
    'abstract class of a statement'
    def __init__(self):
        self.read_expr = []
        self.write_expr = []

class CommentStmt(Statement):
    'comment'
    def __init__(self, comment):
        self.comment = comment

    def __str__(self):
        return '/* %s */' % self.comment

    def copy(self):
        'makes a copy of this object'
        return CommentStmt(self.comment)

class TypeDefStmt(Statement):
    'type definition statement'
    def __init__(self, name, atype, td_type):
        'atype - defined type, td_type - typedef type {typedef,scalarset,...}'
        Statement.__init__(self)
        self.name = name
        self.type = atype
        self.str_type = str(self.type)
        self.type.name = name
        self.td_type = td_type

    def __str__(self):
        return '%s %s %s;' % (self.td_type, id2smv(self.name), self.str_type)

    def copy(self):
        'makes a copy of this object'
        return TypeDefStmt(self.name, self.type, self.td_type)

class ModuleInstStmt(Statement):
    'module instance statement'
    def __init__(self, name, module_name, param_map):
        """instance name, module name, param_map either list of Variables or 
        str->Variable dictionary, where Variable is the object connected to
        input/output port of the module"""
        Statement.__init__(self)
        self.name = name
        self.module_name = module_name
        self.module = None
        # self.param_map after self.unify() will be the map of str->Variable
        # i.e., the map of name_of_port -> object_connected_to_port
        # also, read_expr will be set appropriatelly
        self.param_map = param_map

    def __str__(self):
        if not self.module:
            return '%s : %s(N/A);' % \
                (id2smv(self.name), id2smv(self.module_name))
        if isinstance(self.param_map, list):
            p_str = [str(v) for v in self.param_map]
        else:
            p_str = [str(self.param_map[paramname]) for paramname in
            self.module.paramnames]
#            connected_parameters = self.module.parameters
#            p_str = ['%s<==%s' % (id2smv(port), expr) \
#                for port, expr in self.param_map.items()]
        return '%s: %s(%s);' % \
            (id2smv(self.name), id2smv(self.module.name), ', '.join(p_str))

    def copy(self):
        'makes a copy of this object'
        instance = ModuleInstStmt(self.name, self.module_name,
            self.param_map.copy())
        instance.module = self.module
        return instance

    def unify(self, model):
        """set expected arguments values (self.module, self.param_map)
                1. find the module object in a model
                2. set the parameter map accordingly
        This should be called after the whole SMV source file is parsed, i.e.,
        all modules is known.
        Error: raises TypeError"""
        module = None

        # find the module
        for module in [m for m in model.modules if m.name == self.module_name]:
            break
        if not module:
            raise TypeError('unknown module %s' % self.module_name)
        self.module = module

        # build param_map if the parameter map is a list
        param_map = self.param_map
        if len(param_map) != len(module.paramnames):
            raise TypeError("At module instance: %s\n" \
              "incompatible parameter list for module '%s'" % \
              (self, module.name))
        if isinstance(param_map, list):
            param_map = dict(zip(module.paramnames, param_map))

        # make the param_map of real objects
        self.param_map = {}
        for key, value in param_map.items():
            if not self.module.parameters[key].type.compatible(value.type):
                raise TypeError( \
                    "incompatible type of '%s(%s)' for parameter '%s(%s)'" % \
                    (value, value.type, self.module.parameters[key], \
                    self.module.parameters[key].type))
            self.param_map[key] = value

        # set the list of read expression
        self.read_expr = param_map.values()

class VariableDeclStmt(Statement):
    'variable declaration statement'
    def __init__(self, variable):
        Statement.__init__(self)
        self.variable = variable

    def __str__(self):
        return '%s: %s;' % (self.variable, self.variable.type)

    def copy(self):
        'makes a copy of this object'
        return VariableDeclStmt(self.variable)

class ModParamDeclStmt(Statement):
    'module parameter declaration statement'
    def __init__(self, parameter):
        Statement.__init__(self)
        self.parameter = parameter

    def __str__(self):
        return '%s %s: %s;' % (self.parameter.direction, \
            self.parameter, self.parameter.type)

    def copy(self):
        'makes a copy of this object'
        return ModParamDeclStmt(self.parameter)

class AssignmentStmt(Statement):
    'assignment statement'
    def __init__(self, dst, src, mkaccess=True):
        Statement.__init__(self)
        if mkaccess:
            self.dst = dst.writing()
            self.src = src.reading()
        else:
            self.dst = dst
            self.src = src

        self.read_expr = [src]
        self.write_expr = [dst]

    def __str__(self):
        return '%s := %s;' % (self.dst, self.src)
    
    def copy(self):
        'makes a copy of this object'
        return AssignmentStmt(self.dst, self.src, False)

class IfThenElseStmt(Statement):
    'if-then-else statement'
    def __init__(self, cond, then_stmt, else_stmt=None, mkaccess=True):
        Statement.__init__(self)
        self.cond = cond.reading() if mkaccess else cond
        self.then_stmt = then_stmt
        self.else_stmt = else_stmt

        # set read and written expressions
        self.read_expr = [self.cond] + then_stmt.read_expr
        self.write_expr = then_stmt.write_expr[:]
        if else_stmt:
            self.read_expr += else_stmt.read_expr
            self.write_expr += else_stmt.write_expr

    def __str__(self):
        if_str = 'if (%s)\n%s' % (self.cond, stmt_list2str(self.then_stmt))
        if self.else_stmt:
            if_str += '\nelse\n%s' % stmt_list2str(self.else_stmt)
#        import method3
#        c = method3.find_clocks(self.cond)
#        if c:
#            if_str += '\n'
#            if_str += '----------------------------------------\n'
#            if_str += 'sensitive: %s\n' % ', '.join(['(%s,%s)' % (str(v), t) for
#            v,t in c])
#            if_str += '----------------------------------------\n'
        return if_str
    
    def copy(self):
        'makes a copy of this object'
        return IfThenElseStmt(self.cond, self.then_stmt, self.else_stmt, False)

class DoStmt(Statement):
    'sequential statements'
    def __init__(self, stmt_list):
        Statement.__init__(self)
        self.stmt_list = stmt_list

        # set read and written expressions wrt. statements within DoStmt
        for stmt in stmt_list:
            self.read_expr += stmt.read_expr
            self.write_expr += stmt.write_expr

    def __str__(self):
        return 'do {\n%s\n}' % stmt_list2str(self.stmt_list)
    
    def copy(self):
        'makes a copy of this object'
        return DoStmt(self.stmt_list)

class StmtListStmt(Statement):
    'list of module statements'
    def __init__(self, stmt_list):
        Statement.__init__(self)
        self.stmt_list = stmt_list

        # set read and written expressions wrt. statements within DoStmt
        for stmt in stmt_list:
            self.read_expr += stmt.read_expr
            self.write_expr += stmt.write_expr

    def __str__(self):
        return '{\n%s\n}' % stmt_list2str(self.stmt_list)
    
    def copy(self):
        'makes a copy of this object'
        return StmtListStmt(self.stmt_list)

class AssertionStmt(Statement):
    'assertion class'
    def __init__(self, name, temp_expr, asrtype):
        Statement.__init__(self)
        self.name = name
        self.temp_expr = temp_expr
        self.asrtype = asrtype

    def __str__(self):
        return '%s : %s %s;' % (id2smv(self.name), self.asrtype,
            self.temp_expr)

    def copy(self):
        'makes a copy of this object'
        return AssertionStmt(self.name, self.temp_expr, self.asrtype)

class AssumptionStmt(Statement):
    'assumptions of the model behavior'
    def __init__(self, assump_list):
        Statement.__init__(self)
        self.assump_list = assump_list

    def __str__(self):
        return 'assume %s;' % ','.join([id2smv(a) for a in self.assump_list])

    def copy(self):
        'makes a copy of this object'
        return AssumptionStmt(self.assump_list)

class ProveStmt(Statement):
    'statement saying to prove something'
    def __init__(self, asserts, claim):
        Statement.__init__(self)
        self.asserts = asserts
        self.claim = claim

    def __str__(self):
        return 'using %s prove %s;' % \
            (','.join([id2smv(a) for a in self.asserts]), self.claim)

    def copy(self):
        'makes a copy of this object'
        return ProveStmt(self.asserts, self.claim)

class Expr(object):
    'Abstract class of expressions'
    def __init__(self):
        self.type = None
    
    def __str__(self):
        raise NotImplementedError('internal error: unspecified expression')
    
    def reading(self):
        """returns itself cause it's the real object already. Different
        behavior has a VariableAccess object"""
        return self

    def writing(self):
        """returns itself cause it's the real object already. Different
        behavior has a VariableAccess object"""
        return self
            
class VariableAccess(Expr):
    'variable use'
    def __init__(self, variable, access, lineno):
        Expr.__init__(self)
        self.var = variable
        self.type = variable.type
        self.access = access # 'r'/'w' ('u' - undefined)
        self.lineno = lineno
    
    def __str__(self):
        return str(self.var)
    
    def reading(self):
        'returns the variable read via this access'
        return self.real_object('r')

    def writing(self):
        'returns the variable read via this access'
        return self.real_object('w')
    
    def real_object(self, access):
        """set the access type, add this access to the list of variable
        accesses, and returns the variable itself"""
        self.access = access
        self.var.access.append(self)
        # if it is an array, create the variable access for every of its items
        if isinstance(self.type, ArrayType):
            for variable in self.var.data:
                va = VariableAccess(variable, access, self.lineno)
                va.real_object(access)
        return self.var

class Constant(Expr):
    'a constant within expression'
    def __init__(self, value, ctype):
        Expr.__init__(self)
        self.value = value # an integer or a strings as an id of set item
        self.type = ctype

    def __int__(self):
        if self.type.compatible(INT):
            return self.value
        else:
            raise TypeError('unable to convert %s to integer' % self.value)
    
    def __str__(self):
        return str(self.value)

    def writing(self):
        'causes an error - it is a constant'
        raise SyntaxError('unable to write to a constant')

class UnaryOp(Expr):
    'expression as a unary operator'
    def __init__(self, operator, expr, applyaccess=True):
        Expr.__init__(self)
        self.operator = operator
        if applyaccess:
            expr = expr.reading()
        self.expr = expr
        self.type = expr.type
    
    def __str__(self):
        return '(%s%s)' % (self.operator, str(self.expr))

class Operator(Expr):
    'expression as a non-unary operator'
    def __init__(self, operator, expr_list, applyaccess=True):
        Expr.__init__(self)
        self.operator = operator
        if applyaccess:
            expr_list = [e.reading() for e in expr_list]
        
#        # make the expression be simpler
#        # if the operator is associative,
#        if operator in ('+', '-', '*', '/', '&', '|', '^'):
#            # ...then substitute every operand with the same operator with its
#            # own expr_list
#            # here we make it incrementally, starting with []
#            self.expr_list = []
#            for expr in expr_list:
#                # if it is the same operator expression
#                if isinstance(expr, Operator) and expr.operator == operator:
#                    # add its operands instead...
#                    self.expr_list += expr.expr_list
#                else:
#                    # ...of the expression itself
#                    self.expr_list.append(expr)
#        else:
#            # if it is not associative, make the operands unchanged
#            self.expr_list = expr_list
        self.expr_list = expr_list
        
        self.typecheck()

    def __str__(self):
        return '(%s)' % (' '+self.operator+' ').join( \
            [str(e) for e in self.expr_list])

    def typecheck(self):
        'perform typechecking and set the self.type; raises TypeError'
        # arithmetic operators expect IntType
        if self.operator in ('+', '-', '*'):
            for expr in self.expr_list:
                if not isinstance(expr.type, IntType):
                    raise TypeError("'%s(%s)' must be integer type" % \
                        (expr, expr.type))
            self.type = INT

        # logic operators expect boolean type or array of booleans
        elif self.operator in ('&', '|', '^'):
            etype = self.expr_list[0].type
            if not BOOLEAN.compatible(etype) and \
               not BOOLARRAY.compatible(etype):
                raise TypeError( \
                    "'%s(%s)' must be either boolean or array of booleans" % \
                    (self.expr_list[0], self.expr_list[0].type))
            for expr in self.expr_list[1:]:
                if not etype.compatible(expr.type):
                    raise TypeError( \
                        "'%s(%s)' must be of type '%s'" % \
                            (expr, expr.type, etype))
            self.type = etype

        # relational operators expect the same type of all the operands
        elif self.operator in ('=', '!=', '<', '<=', '>', '>='):
            etype = self.expr_list[0].type
            for expr in self.expr_list[1:]:
                if not etype.compatible(expr.type):
                    raise TypeError( \
                        "'%s(%s)' must be of type '%s'" % \
                            (expr, expr.type, etype))
            self.type = BOOLEAN

        else:
            raise TypeError("unknown operator '%s'" % self.operator)

class TernaryOp(Expr):
    'expression using ternary operator ? :'
    def __init__(self, cond, expr1, expr2, applyaccess=True):
        Expr.__init__(self)

        # type check
        if not cond.type.compatible(BOOLEAN):
            raise TypeError("incompatible type '%s'" % cond.type)
        self.type = expr1.type
        if not expr1.type.compatible(expr2.type):
            raise TypeError("incompatible types '%s(%s)' and '%s(%s)'" % \
                (expr1.type, expr1, expr2.type, expr2))

        # self.data assignment
        if applyaccess:
            self.data = [cond.reading(), expr1.reading(), expr2.reading()]
        else:
            self.data = [cond, expr1, expr2]

    def __str__(self):
        return '(%s) ? (%s) : (%s)' % self.data

class ArrayCat(Expr):
    'array concatenation represented by :: operator'
    def __init__(self, left, right):
        # one (or possibly both) left and right might be None, which means that
        # no left or right part will be appended to data - suitable for making
        # a copy of ArrayCat.
        Expr.__init__(self)
        self.data = []
        self.type = ArrayType('', 0, 0, BOOLEAN)
        for part in (left, right):
            if part:
                self.append(part.reading())

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return '::'.join([str(d) for d in self.data])

    def append(self, instance):
        'append an instance to the array'
        if instance.type.compatible(BOOLEAN):
            self.data.append(instance)
            self.type.upper += 1
        elif instance.type.compatible(BOOLARRAY):
            self.data += instance.data
            self.type.upper += len(instance)
        else:
            # type check failed (only boolean base class is supported)
            raise TypeError('%s must be boolean or array of booleans' % \
                instance)

    def import_data(self, data):
        'import a data and set the size of the array appropriatelly'
        self.data = data
        self.type.upper = len(data)

class TempOp(Expr):
    'temporal init/next(expr) operator'
    def __init__(self, expr):
        Expr.__init__(self)
        self.expr = expr
        self.type = expr.type

    def __str__(self):
        raise NotImplementedError( \
            'internal error: uninitialized temporal operator')

    def reading(self):
        'returns the read expression'
        self.expr = self.expr.reading()
        return self

    def writing(self):
        'returns the written expression'
        self.expr = self.expr.writing()
        return self
    
class InitOp(TempOp):
    'temporal init(epxr) operator'
    def __str__(self):
        return 'init(%s)' % self.expr
    
class NextOp(TempOp):
    'temporal next(epxr) operator'
    def __str__(self):
        return 'next(%s)' % self.expr

class SwitchOp(Expr):
    'switch(expr){cases} expression'
    def __init__(self, expr, cases, applyaccess=True):
        """cases is expected to be the list of tuples (e1, e2); e1 compatible
        with cond"""
        assert isinstance(cases, list)
        assert len(cases)
        Expr.__init__(self)
        if applyaccess:
            expr = expr.reading()
        self.cond = expr
        self.type = cases[0][1].type
        # type check
        for expr1, expr2 in cases:
            if not expr1.type.compatible(self.cond.type):
                raise TypeError('incompatible types %s and %s' % \
                    (expr, expr1))
            if not expr2.type.compatible(self.type):
                raise TypeError('incompatible types among swith cases')
        if applyaccess:
            cases = [(e1.reading(), e2.reading()) for e1, e2 in cases]
        self.cases = cases

    def __str__(self):
        return 'switch(%s) {\n%s\n}' % (self.cond, \
            '\n'.join(['    %s : %s;' % case for case in self.cases]))

class CaseBlock(Expr):
    'case { (case : expr;)+ }'
    def __init__(self, cases, applyaccess=True):
        """cases is the list of tuples (cond, e); cond must be compatible with
        boolean, every e must be compatible with each other"""
        assert isinstance(cases, list)
        assert len(cases) > 0
        Expr.__init__(self)
        # type check
        self.type = cases[0][1].type
        for cond, expr in cases:
            if not cond.type.compatible(BOOLEAN):
                raise TypeError('%s must be boolean' % cond)
            if not expr.type.compatible(self.type):
                raise TypeError('incompatible types in different case blocks')
        if applyaccess:
            cases = [(cond.reading(), e.reading()) for cond, e in cases]
        self.cases = cases

    def __str__(self):
        return 'case {\n%s\n}' % \
            '\n'.join(['  %s : %s;' % case for case in self.cases])

def copy_expr(expr, var_map):
    """return a new syntax tree for the expression; every link to some variable
    or parameter will be replaced by the one from the variable_map
    (str->Variable = name of variable or parameter -> variable object).
    This function should be called on an expression of operators, variables,
    parameters, and constants. VariableAccess is not supported."""

    # Variable returns new variable from var_map
    if isinstance(expr, Variable):
        # if variable is in the var_map, returns its representative
        if expr.name in var_map.keys():
            return var_map[expr.name]
        # else variable itself (not found in the var_map)
        else:
            return expr

    # Constant returns itself
    elif isinstance(expr, Constant):
        return expr

    # UnaryOp returns new UnaryOp with copied expr
    elif isinstance(expr, UnaryOp):
        return UnaryOp(expr.operator, copy_expr(expr.expr, var_map), False)

    # Operator returns new Operator with copied expr_list
    elif isinstance(expr, Operator):
        return Operator(expr.operator, 
            [copy_expr(e, var_map) for e in expr.expr_list], False)

    # TernaryOp returns new TernaryOp with copied expressions
    elif isinstance(expr, TernaryOp):
        return TernaryOp(
            copy_expr(expr.data[0], var_map),
            copy_expr(expr.data[1], var_map),
            copy_expr(expr.data[2], var_map),
            False)

    # ArrayCat returns new ArrayCat of copyied parts
    elif isinstance(expr, ArrayCat):
        # make ArrayCat of: [copied_data]::None
        array_cat_copy = ArrayCat(None, None)
        array_cat_copy.import_data([copy_expr(d, var_map) for d in expr.data])
        return array_cat_copy

    # TempOp returns new TempOp of copyied expression
    elif isinstance(expr, TempOp):
        return type(expr)(copy_expr(expr.expr, var_map))

    # SwitchOp returns new SwitchOp with copied expr and cases
    elif isinstance(expr, SwitchOp):
        cases = [(copy_expr(e1, var_map), copy_expr(e2, var_map))
            for e1, e2 in expr.cases]
        return SwitchOp(copy_expr(expr.cond, var_map), cases, False)

    # CaseBlock returns new CaseBlock with copied cases
    elif isinstance(expr, CaseBlock):
        cases = [(copy_expr(cond, var_map), copy_expr(e, var_map))
            for cond, e in expr.cases]
        return CaseBlock(cases, False)

    # other types are unsupported
    else:
        TypeError("unable to copy expression of type %s", type(expr))

def copy_stmt(stmt, var_map):
    '''makes copied version of a statement
    Should be called by mkinstance() and recursivelly by itself.
    Raises TypeError if unknown statement type is found'''

    # copy the statement
    nstmt = stmt.copy()

    # perform some changes, if required:

    if isinstance(stmt, ModuleInstStmt):
        # substitute every expression assigned to a moduel parameter by
        # its copy containing the variables and parameters of var_map
        for name, expr in nstmt.param_map.items():
            nstmt.param_map[name] = copy_expr(expr, var_map)
        # and because we are now making an instance of this module, we
        # also make an instance of such a submodule accordingly
        #nstmt.module = stmt.module.mkinstance()

    # VariableDeclStmt will point to the new variable
    elif isinstance(stmt, VariableDeclStmt):
        # find the new variable in already existing name->Variable map
        # make a new variable declaration pointing to the new variable
        nstmt.variable = var_map[stmt.variable.name]

    # ModParamDeclStmt will point to the new module parameter (similar
    # to VariableDeclStmt handling)
    elif isinstance(stmt, ModParamDeclStmt):
        # find the new parameter
        nstmt.parameter = var_map[stmt.parameter.name]

    # AssignmentStmt with copied expressions
    elif isinstance(stmt, AssignmentStmt):
        nstmt.dst = copy_expr(stmt.dst, var_map)
        nstmt.src = copy_expr(stmt.src, var_map)

    # IfThenElseStmt with copied condition and all statements
    elif isinstance(stmt, IfThenElseStmt):
        nstmt.cond = copy_expr(stmt.cond, var_map)
        nstmt.then_stmt = copy_stmt(stmt.then_stmt, var_map)
        if nstmt.else_stmt:
            nstmt.else_stmt = copy_stmt(stmt.else_stmt, var_map)

    # DoStmt and StmtListStmt makes every of their statement being copied
    elif isinstance(stmt, DoStmt) or isinstance(stmt, StmtListStmt):
        nstmt.stmt_list = [copy_stmt(s, var_map) for s in nstmt.stmt_list]

    return nstmt

INDENT_LEVEL = '    '

def stmt_list2str(stmt_list, indent=INDENT_LEVEL):
    'returns statements in SMV language prefixed with appropriate level'
    if not isinstance(stmt_list, list):
        stmt_list = [stmt_list]
    lines = []
    for stmt in stmt_list:
        lines += str(stmt).split('\n')
    string_stmt = '\n'.join([indent+line for line in lines])
    return string_stmt

SMV_RESERVED_TOKENS = (
    'TYPEDEF', 'SCALARSET', 'ORDSET', 'MODULE', 'INPUT', 'OUTPUT',
    'BOOLEAN', 'RESOLVE', 'IF', 'ELSE', 'INIT', 'NEXT', 'DO', 'CASE',
    'SWITCH', 'ARRAY', 'OF', 'UNDEFINED', 'ASSERT', 'ASSUME', 'USING',
    'PROVE', 'SPEC', 'IN',
)
NORMAL_ID_REGEXP = re.compile('^[a-z0-9_ ]*$')
def id2smv(ident):
    'returns string identifier conforming to SMV syntax'
    return ident \
        if NORMAL_ID_REGEXP.match(ident) and \
           ident.upper() not in SMV_RESERVED_TOKENS \
        else '\\'+ident+' '
