#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""smvparser.py - SMV language parser"""

import ply.lex
import ply.yacc
from model import *

class SMVError(Exception):
    'Exception for a parser'
    pass

class AbstractParser(object):
    'Base class for a lexer/parser'
    tokens = ()
    precedence = ()

    def __init__(self):
        self.lexer = ply.lex.lex(module=self)
        self.parser = ply.yacc.yacc(module=self)
        self.filename = ''

    def parse(self, filename):
        "Read file's content and parse it"
        self.filename = filename
        smvfile = open(self.filename)
        content = smvfile.read()
        smvfile.close()
        ply.yacc.parse(content)

class Parser(AbstractParser):
    'Parser class for a subset of Cadence SMV language'
    
    # Tokens
    tokens_reserved = SMV_RESERVED_TOKENS

    tokens_tempop = ('EX', 'AX', 'EF', 'AF', 'EG', 'AG', 'E', 'A', 'X',
        'F', 'G', 'U', 'W',
    )

    tokens_symbol = (
        'SEMI', 'COLON', 'COMMA', 'LPAREN', 'RPAREN', 'LBRACE', 'RBRACE',
        'LBRACKET', 'RBRACKET', 'CONDOP', 'TO', 'FOURDOT', 'CONNECT', 'ASSIGN',
        'PLUS', 'MINUS', 'TIMES', 'AND', 'OR', 'XOR', 'NOT', 'LNOT', 'IMPLY',
        'EQ', 'NEQ', 'LT', 'LE', 'GT', 'GE',
    )

    tokens = tokens_reserved + tokens_tempop + tokens_symbol + ('ID', 'NUMBER')

    t_ignore    = ' \t'
    t_SEMI      = r';'
    t_COLON     = r':'
    t_COMMA     = r','
    t_LPAREN    = r'\('
    t_RPAREN    = r'\)'
    t_LBRACE    = r'\{'
    t_RBRACE    = r'\}'
    t_LBRACKET  = r'\['
    t_RBRACKET  = r'\]'
    t_CONDOP    = r'\?'
    t_TO        = r'\.\.'
    t_FOURDOT   = r'::'
    t_CONNECT   = r'<=='
    t_ASSIGN    = r':='
    t_PLUS      = r'\+'
    t_MINUS     = r'-'
    t_TIMES     = r'\*'
    t_AND       = r'&'
    t_OR        = r'\|'
    t_XOR       = r'\^'
    t_NOT       = r'~'
    t_LNOT      = r'!'
    t_EQ        = r'='
    t_NEQ       = r'!='
    t_LT        = r'<'
    t_LE        = r'<='
    t_GT        = r'>'
    t_GE        = r'>='
    t_IMPLY     = r'->'
    t_NUMBER    = r'\d+'

    reserved_map = {}

    def __init__(self, model):
        self.model = model
        for token in self.tokens_reserved:
            self.reserved_map[token.lower()] = token
        for token in self.tokens_tempop:
            self.reserved_map[token] = token
        AbstractParser.__init__(self)

    def parse(self, filename):
        'parse the file and fill up the model'
        # parse the file
        AbstractParser.parse(self, filename)
        # !!! try this after debugging 
#        try:
#            AbstractParser.parse(self, filename)
#        except TypeError, err:
#            self.error(err.message)

        # unify all module instances, i.e., make them be link to modules, not
        # only the links to module names
        for module in self.model.modules:
            for minstance in [stmt for stmt in module.stmts \
                if type(stmt) == ModuleInstStmt]:
                minstance.unify(self.model)

        # !!! DELETE THIS
        # make every of module instance be really the instance, not only the
        # link to a module
#        for module in self.model.modules:
#            for minstance in [stmt for stmt in module.stmts \
#                if type(stmt) == ModuleInstStmt]:
#                minstance.module = minstance.module.mkinstance()

    def t_id(self, tok):
        r'([a-zA-Z$_][a-zA-Z0-9$_]*|\\[^ \n]* )'
        if tok.value in self.tokens_tempop:
            tok.type = tok.value
        else:
            tok.type = self.reserved_map.get(tok.value.lower(), 'ID')
        if tok.value[0] == '\\':
            tok.value = tok.value[1:]
        return tok

    def t_newline(self, tok):
        r'\n+'
        self.lexer.lineno += tok.value.count('\n')

    def t_comment(self, tok):
        r'(/\*(.|\n)*?\*/|--.*\n)'
        self.lexer.lineno += tok.value.count('\n')

    def t_preprocessor(self, tok):
        r'\#(.)*?\n'
        tok.lexer.lineno += 1 # better self.lexer, but pylint warns

    def t_error(self, tok):
        'when lexer error occures'
        self.error("illegal character '%s'" % tok.value[0])

    # Parsing rules
    def p_start(self, tok):
        'start : prog_stmt_list'
    def p_prog_stmt_list1(self, tok):
        'prog_stmt_list : prog_stmt'
        self.model.stmts.append(tok[1])
    def p_prog_stmt_list2(self, tok):
        'prog_stmt_list : prog_stmt_list prog_stmt'
        self.model.stmts.append(tok[2])
    def p_prog_stmt1(self, tok):
        'prog_stmt : type_def SEMI'
        tok[0] = tok[1]
    def p_prog_stmt2(self, tok):
        'prog_stmt : module'
        tok[0] = tok[1]
    def p_type_def1(self, tok):
        'type_def : TYPEDEF ID type'
        tok[0] = TypeDefStmt(tok[2], tok[3], 'typedef')
    def p_type_def2(self, tok):
        'type_def : SCALARSET ID expr TO expr'
        tok[0] = TypeDefStmt(tok[2], IntType(tok[3], tok[5]), 'scalarset')
    def p_type_def3(self, tok):
        'type_def : ORDSET ID expr TO'
        tok[0] = TypeDefStmt(tok[2], IntType(tok[3], ''), 'ordset')
    def p_module(self, tok):
        'module : MODULE ID LPAREN id_list_opt module1 RPAREN module_body'
        self.model.modules[-1].stmts = tok[7]
        tok[0] = self.model.modules[-1]
    def p_module1(self, tok):
        'module1 :'
        self.model.modules.append(Module(tok[-3], tok[-1]))
    def p_module_body(self, tok):
        'module_body : LBRACE stmt_list RBRACE'
        tok[0] = tok[2]
    def p_id_list_opt1(self, tok):
        'id_list_opt : empty'
        tok[0] = []
    def p_id_list_opt2(self, tok):
        'id_list_opt : id_list'
        tok[0] = tok[1]
    def p_id_list1(self, tok):
        'id_list : ID'
        tok[0] = [tok[1]]
    def p_id_list2(self, tok):
        'id_list : id_list COMMA ID'
        tok[0] = tok[1] + [tok[3]]
    def p_stmt_list1(self, tok):
        'stmt_list : stmt'
        tok[0] = tok[1] if isinstance(tok[1], list) else [tok[1]]
    def p_stmt_list2(self, tok):
        'stmt_list : stmt_list stmt'
        if isinstance(tok[2], list):
            tok[0] = tok[1] + tok[2]
        else:
            tok[0] = tok[1] + [tok[2]]
    def p_stmt1(self, tok):
        'stmt : var_decl SEMI'
        tok[0] = tok[1]
    def p_stmt2(self, tok):
        'stmt : type_def SEMI'
        tok[0] = tok[1]
    def p_stmt3(self, tok):
        'stmt : assignment SEMI'
        tok[0] = tok[1]
    def p_stmt4(self, tok):
        'stmt : module_def SEMI'
        tok[0] = tok[1]
    def p_stmt5(self, tok):
        'stmt : if_stmt'
        tok[0] = tok[1]
    def p_stmt6(self, tok):
        'stmt : DO block'
        tok[0] = DoStmt(tok[2])
    def p_stmt7(self, tok):
        'stmt : block'
        tok[0] = StmtListStmt(tok[1])
    def p_stmt8(self, tok):
        'stmt : SEMI'
        tok[0] = ';'
    def p_stmt9(self, tok):
        'stmt : assertion SEMI'
        tok[0] = tok[1]
    def p_stmt10(self, tok):
        'stmt : assumption SEMI'
        tok[0] = tok[1]
    def p_stmt11(self, tok):
        'stmt : prove SEMI'
        tok[0] = tok[1]
    def p_assertion1(self, tok):
        'assertion : id_list COLON ASSERT te'
        if len(tok[1]) > 1:
            self.error( \
                'unexpected list of identifiers in an assertion',
                tok.lineno(2))
        tok[0] = AssertionStmt(tok[1][0], tok[4], 'assert')
    def p_assertion2(self, tok):
        'assertion : id_list COLON SPEC te'
        if len(tok[1]) > 1:
            self.error( \
                'unexpected list of identifiers in a specification',
                tok.lineno(2))
        tok[0] = AssertionStmt(tok[1][0], tok[4], 'SPEC')
    def p_assumption(self, tok):
        'assumption : ASSUME id_list'
        tok[0] = AssumptionStmt(tok[2])
    def p_prove(self, tok):
        'prove : USING id_list PROVE ID'
        tok[0] = ProveStmt(tok[2], tok[4])
    def p_block(self, tok):
        'block : LBRACE stmt_list RBRACE'
        tok[0] = tok[2]
    def p_var_decl1(self, tok):
        'var_decl : INPUT id_list COLON type'
        # zaradit do Module.parameters
        assert type(tok[2]) == list
        assert isinstance(tok[4], AbstractType)
        decls = []
        for ident in tok[2]:
            param = ModuleParameter(ident, tok[4], 'input')
            self.model.modules[-1].parameters[ident] = param
            decls.append(ModParamDeclStmt(param))
        tok[0] = decls
    def p_var_decl2(self, tok):
        'var_decl : OUTPUT id_list COLON type'
        # zaradit do Module.parameters
        assert type(tok[2]) == list
        assert isinstance(tok[4], AbstractType)
        decls = []
        for ident in tok[2]:
            param = ModuleParameter(ident, tok[4], 'output')
            self.model.modules[-1].parameters[ident] = param
            decls.append(ModParamDeclStmt(param))
        tok[0] = decls
    def p_var_decl3(self, tok):
        'var_decl : id_list COLON type'
        # zaradit do Module.variables
        assert type(tok[1]) == list
        assert isinstance(tok[3], AbstractType)
        decls = []
        for ident in tok[1]:
            var = Variable(ident, tok[3])
            self.model.modules[-1].variables[ident] = var
            decls.append(VariableDeclStmt(var))
        tok[0] = decls
    def p_assignment(self, tok):
        'assignment : temp_value ASSIGN expr'
        tok[0] = AssignmentStmt(tok[1], tok[3])
    def p_module_def(self, tok):
        'module_def : id_list COLON ID LPAREN param_map_list_opt RPAREN'
        assert type(tok[1]) == list
        # id_list must have 1 item only; it is id_list due to grammar conflict
        if len(tok[1]) > 1:
            self.error( \
                'unexpected list of identifiers in module instantiation',
                tok.lineno(2))
        tok[0] = ModuleInstStmt(tok[1][0], tok[3], tok[5])
    def p_param_map_list_opt1(self, tok):
        'param_map_list_opt : empty'
        tok[0] = []
    def p_param_map_list_opt2(self, tok):
        'param_map_list_opt : read_object_list'
        tok[0] = tok[1]
    def p_param_map_list_opt3(self, tok):
        'param_map_list_opt : param_map_list'
        tok[0] = tok[1]
    def p_read_object_list(self, tok):
        'read_object_list : object_list'
        tok[0] = [obj.reading() for obj in tok[1]]
    def p_object_list1(self, tok):
        'object_list : object'
        tok[0] = [tok[1]]
    def p_object_list2(self, tok):
        'object_list : object_list COMMA object'
        tok[0] = tok[1] + [tok[3]]
    def p_param_map_list1(self, tok):
        'param_map_list : param_map'
        key, value = tok[1]
        tok[0] = {key: value}
    def p_param_map_list2(self, tok):
        'param_map_list : param_map_list COMMA param_map'
        key, value = tok[3]
        tok[1][key] = value
        tok[0] = tok[1]
    def p_param_map(self, tok):
        'param_map : ID CONNECT read_expr'
        tok[0] = (tok[1], tok[3])
    def p_if_stmt1(self, tok):
        'if_stmt : IF LPAREN expr RPAREN stmt'
        tok[0] = IfThenElseStmt(tok[3], tok[5])
    def p_if_stmt2(self, tok):
        'if_stmt : IF LPAREN expr RPAREN stmt ELSE stmt'
        tok[0] = IfThenElseStmt(tok[3], tok[5], tok[7])
    def p_type1(self, tok):
        'type : BOOLEAN'
        tok[0] = BooleanType()
    def p_type2(self, tok):
        'type : BOOLEAN RESOLVE'
        tok[0] = BooleanType(name='boolean resolve')
    def p_type3(self, tok):
        'type : ARRAY expr TO expr OF type'
        tok[0] = ArrayType('', tok[2], tok[4], tok[6])
        tok[0].lineno = tok.lineno(1)
    def p_type4(self, tok):
        'type : expr TO expr'
        tok[0] = IntType(tok[1], tok[3])
        tok[0].lineno = tok.lineno(2)
    def p_type5(self, tok):
        'type : set_def'
        tok[0] = tok[1]
    def p_type6(self, tok):
        'type : ID'
        obj = self.model.find_id(tok[1])
        if isinstance(obj, TypeDefStmt):
            tok[0] = obj.type
        else:
            self.error('%s must be type identifier' % tok[1], tok.lineno(1))
    def p_set_def(self, tok):
        'set_def : LBRACE primary_list RBRACE'
        plist = tok[2]
        first_type = type(plist[0])
        for item in plist:
            if type(item) != first_type:
                self.error('unexpected combination of %s and %s' %  \
                    (plist[0], item), tok.lineno(1))
        tok[0] = SetType(plist)
        tok[0].lineno = tok.lineno(1)
    def p_primary_list1(self, tok):
        'primary_list : number_list'
        tok[0] = tok[1]
    def p_primary_list2(self, tok):
        'primary_list : id_list'
        tok[0] = tok[1]
    def p_number_list1(self, tok):
        'number_list : NUMBER'
        tok[0] = [int(tok[1])]
    def p_number_list2(self, tok):
        'number_list : number_list COMMA NUMBER'
        tok[0] = tok[1] + [int(tok[3])]
    def p_read_expr(self, tok):
        'read_expr : expr'
        # if expr is Variable reference, create VariableAccess('r')
        try:
            tok[0] = tok[1].reading()
        except:
            print self.lexer.lineno
    def p_expr1(self, tok):
        'expr : expr_item'
        tok[0] = tok[1]
    def p_expr2(self, tok):
        'expr : expr FOURDOT expr_item'
        try:
            tok[0] = ArrayCat(tok[1], tok[3])
            tok[0].lineno = tok.lineno(2)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr3(self, tok):
        'expr : SWITCH LPAREN object RPAREN LBRACE swblock RBRACE'
        try:
            tok[0] =  SwitchOp(tok[3], tok[6])
            tok[0].lineno = tok.lineno(1)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr4(self, tok):
        'expr : CASE LBRACE cblock RBRACE'
        try:
            tok[0] = CaseBlock(tok[3])
            tok[0].lineno = tok.lineno(1)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr5(self, tok):
        'expr : expr CONDOP expr COLON expr'
        try:
            tok[0] = TernaryOp(tok[1], tok[3], tok[5])
            tok[0].lineno = tok.lineno(2)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_swblock1(self, tok):
        'swblock : swstmt'
        tok[0] = [tok[1]]
    def p_swblock2(self, tok):
        'swblock : swblock swstmt'
        tok[0] = tok[1] + [tok[2]]
    def p_swstmt(self, tok):
        'swstmt : expr COLON expr SEMI'
        tok[0] = (tok[1], tok[3])
    def p_cblock1(self, tok):
        'cblock : cstmt'
        tok[0] = [tok[1]]
    def p_cblock2(self, tok):
        'cblock : cblock cstmt'
        tok[0] = tok[1] + [tok[2]]
    def p_cstmt(self, tok):
        'cstmt : expr_rel COLON expr SEMI'
        tok[0] = (tok[1], tok[3])
    def p_expr_item1(self, tok):
        'expr_item : expr_rel'
        tok[0] = tok[1]
    def p_expr_rel1(self, tok):
        'expr_rel : expr_log1'
        tok[0] = tok[1]
    def p_expr_rel2(self, tok):
        'expr_rel : expr_rel rel_op expr_log1'
        try:
            tok[0] = Operator(tok[2], [tok[1], tok[3]])
            tok[0].lineno = tok[1].lineno
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_rel_op(self, tok):
        """rel_op : EQ
                  | NEQ
                  | LT
                  | LE
                  | GT
                  | GE"""
        tok[0] = tok[1]
    def p_expr_log1a(self, tok):
        'expr_log1 : expr_log2'
        tok[0] = tok[1]
    def p_expr_log1b(self, tok):
        'expr_log1 : expr_log1 OR expr_log2'
        try:
            tok[0] = Operator(tok[2], [tok[1], tok[3]])
            tok[0].lineno = tok.lineno(2)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr_log1c(self, tok):
        'expr_log1 : expr_log1 XOR expr_log2'
        try:
            tok[0] = Operator(tok[2], [tok[1], tok[3]])
            tok[0].lineno = tok.lineno(2)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr_log2d(self, tok):
        'expr_log2 : expr_arith'
        tok[0] = tok[1]
    def p_expr_log2e(self, tok):
        'expr_log2 : expr_log2 AND expr_arith'
        try:
            tok[0] = Operator(tok[2], [tok[1], tok[3]])
            tok[0].lineno = tok.lineno(2)
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_expr_arith1(self, tok):
        'expr_arith : expr_simple'
        tok[0] = tok[1]
    def p_expr_arith2(self, tok):
        'expr_arith : expr_arith arith_op expr_simple'
        try:
            tok[0] = Operator(tok[2], [tok[1], tok[3]])
            tok[0].lineno = tok[1].lineno
        except TypeError, err:
            self.error(err.message, tok.lineno(2))
    def p_arith_op(self, tok):
        """arith_op : PLUS
                    | MINUS
                    | TIMES"""
        tok[0] = tok[1]
    def p_expr_simple1(self, tok):
        'expr_simple : expr_set'
        tok[0] = tok[1]
    def p_expr_simple2(self, tok):
        'expr_simple : unary_op expr_simple'
        tok[0] = UnaryOp(tok[1], tok[2])
        tok[0].lineno = tok[2].lineno
    def p_unary_op(self, tok):
        """unary_op : PLUS
                    | MINUS
                    | NOT
                    | LNOT
                    | AND
                    | OR
                    | XOR"""
        tok[0] = tok[1]
    def p_expr_set1(self, tok):
        'expr_set : temp_value'
        tok[0] = tok[1]
    def p_expr_set2(self, tok):
        'expr_set : set_def'
        tok[0] = Constant(str(tok[1]), tok[1])
        tok[0].lineno = tok.lineno(1)
    def p_temp_value1(self, tok):
        'temp_value : expr_index'
        tok[0] = tok[1]
    def p_temp_value2(self, tok):
        'temp_value : INIT LPAREN object RPAREN'
        tok[0] = InitOp(tok[3])
        tok[0].lineno = tok.lineno(1)
    def p_temp_value3(self, tok):
        'temp_value : next_value'
        tok[0] = tok[1]
    def p_next_value4(self, tok):
        'next_value : NEXT LPAREN object RPAREN'
        tok[0] = NextOp(tok[3])
        tok[0].lineno = tok.lineno(1)
    def p_next_value5(self, tok):
        'next_value : NEXT LPAREN next_value RPAREN'
        tok[0] = NextOp(tok[3])
        tok[0].lineno = tok.lineno(1)
    def p_expr_index1(self, tok):
        'expr_index : primary'
        tok[0] = tok[1]
    def p_expr_index2(self, tok):
        'expr_index : expr_index LBRACKET expr RBRACKET'
        if not isinstance(tok[1].type, ArrayType):
            self.error('non-array type of %s' % tok[1], tok.lineno(2))
        if not isinstance(tok[1], VariableAccess):
            self.error('unable to access to an item of %s' % tok[1],
                tok.lineno(2))
        if not isinstance(tok[3], Constant) or \
          not isinstance(tok[3].value, int):
            self.error('unable to enumerate item by %s' % tok[3],
                tok.lineno(4))
        tok[0] = VariableAccess(tok[1].var.get_item(tok[3].value), 'u',
            tok.lineno(2))
               #!!! uprav ziskani n-te polozky (tok[3].value)-te polozky
               # pomoci tok[1].var.get_item()
#        tok[0] = VariableAccess(tok[1].var.data[tok[3].value], 'u',
#            tok.lineno(2))
    def p_primary1(self, tok):
        'primary : NUMBER'
        tok[0] = Constant(int(tok[1]), INT)
        tok[0].lineno = tok.lineno(1)
    def p_primary2(self, tok):
        'primary : ID'
        obj = self.model.find_id(tok[1])
        if isinstance(obj, Constant):
            tok[0] = obj
        elif isinstance(obj, Variable):
            tok[0] = VariableAccess(obj, 'u', tok.lineno(1))
        else:
            self.error("'%s' - variable or constant identifier expected" % \
                tok[1], tok.lineno(1))
    def p_primary3(self, tok):
        'primary : LPAREN expr RPAREN'
        tok[0] = tok[2]
    def p_primary4(self, tok):
        'primary : UNDEFINED'
        tok[0] = Constant('undefined', UNDEFINED)
        tok[0].lineno = tok.lineno(1)
    def p_object1(self, tok):
        'object : ID'
        var = self.model.find_id(tok[1])
        if not isinstance(var, Variable):
            self.error('variable identifier expected %s' % tok[1],
                tok.lineno(1))
        else:
            tok[0] = VariableAccess(var, 'u', tok.lineno(1))
    def p_object2(self, tok):
        'object : object LBRACKET expr RBRACKET'
        if not isinstance(tok[1].type, ArrayType):
            self.error('non-array type of %s' % tok[1], tok.lineno(2))
        if not isinstance(tok[1], VariableAccess):
            self.error('unable to access to an item of %s (%s)' % \
                (tok[1], type(tok[1])), tok.lineno(2))
        if not isinstance(tok[3], Constant) or \
          not isinstance(tok[3].value, int):
            self.error('unable to enumerate item by %s' % tok[3],
                tok.lineno(4))
        tok[0] = VariableAccess(tok[1].var.data[tok[3].value], 'u',
            tok.lineno(1))
    def p_object3(self, tok):
        'object : LPAREN object RPAREN'
        tok[0] = tok[1]
    def p_empty(self, tok):
        'empty :'
    precedence = (
        ('right', 'CONDOP'),
        ('left', 'IMPLY'),
        ('left', 'OR', 'XOR'),
        ('left', 'AND'),
        ('left', 'NOT', 'LNOT'),
        ('left', 'EX', 'AX', 'EF', 'AF', 'EG', 'AG', 'E', 'A', \
                  'X', 'F', 'G', 'U', 'W'),
        ('left', 'EQ', 'NEQ', 'LT', 'GT', 'LE', 'GE'),
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES'),
        ('right', 'UMINUS'),
    )
    def tok2str(self, tok):
        'returns a string simply by joining all tokens'
        result = []
        for index in range(1, len(tok)):
            result.append(tok[index])
        return ' '.join(result)
    def p_te1(self, tok):
        'te : ID'
        tok[0] = self.tok2str(tok)
    def p_te2(self, tok):
        'te : NUMBER'
        tok[0] = str(tok[1])
    def p_te3(self, tok):
        'te : LPAREN te RPAREN'
        tok[0] = self.tok2str(tok)
    def p_te4(self, tok):
        'te : te IMPLY te'
        tok[0] = self.tok2str(tok)
    def p_te5(self, tok):
        'te : te XOR te'
        tok[0] = self.tok2str(tok)
    def p_te6(self, tok):
        'te : te OR te'
        tok[0] = self.tok2str(tok)
    def p_te7(self, tok):
        'te : te AND te'
        tok[0] = self.tok2str(tok)
    def p_te8(self, tok):
        'te : NOT te'
        tok[0] = self.tok2str(tok)
    def p_te9(self, tok):
        'te : EX te'
        tok[0] = self.tok2str(tok)
    def p_te10(self, tok):
        'te : AX te'
        tok[0] = self.tok2str(tok)
    def p_te11(self, tok):
        'te : EF te'
        tok[0] = self.tok2str(tok)
    def p_te12(self, tok):
        'te : AF te'
        tok[0] = self.tok2str(tok)
    def p_te13(self, tok):
        'te : EG te'
        tok[0] = self.tok2str(tok)
    def p_te14(self, tok):
        'te : AG te'
        tok[0] = self.tok2str(tok)
    def p_te15(self, tok):
        'te : E te'
        tok[0] = self.tok2str(tok)
    def p_te16(self, tok):
        'te : A te'
        tok[0] = self.tok2str(tok)
    def p_te17(self, tok):
        'te : X te'
        tok[0] = self.tok2str(tok)
    def p_te18(self, tok):
        'te : F te'
        tok[0] = self.tok2str(tok)
    def p_te19(self, tok):
        'te : G te'
        tok[0] = self.tok2str(tok)
    def p_te20(self, tok):
        'te : te U te'
        tok[0] = self.tok2str(tok)
    def p_te21(self, tok):
        'te : te W te'
        tok[0] = self.tok2str(tok)
    def p_te22(self, tok):
        'te : te EQ te'
        tok[0] = self.tok2str(tok)
    def p_te23(self, tok):
        'te : te NEQ te'
        tok[0] = self.tok2str(tok)
    def p_te24(self, tok):
        'te : te LT te'
        tok[0] = self.tok2str(tok)
    def p_te25(self, tok):
        'te : te GT te'
        tok[0] = self.tok2str(tok)
    def p_te26(self, tok):
        'te : te LE te'
        tok[0] = self.tok2str(tok)
    def p_te27(self, tok):
        'te : te GE te'
        tok[0] = self.tok2str(tok)
    def p_te28(self, tok):
        'te : te PLUS te'
        tok[0] = self.tok2str(tok)
    def p_te29(self, tok):
        'te : te MINUS te'
        tok[0] = self.tok2str(tok)
    def p_te30(self, tok):
        'te : te TIMES te'
        tok[0] = self.tok2str(tok)
    def p_te31(self, tok):
        'te : MINUS te %prec UMINUS'
        tok[0] = self.tok2str(tok)
    def p_te32(self, tok):
        'te : LNOT te'
        tok[0] = self.tok2str(tok)

    def p_error(self, tok):
        'when parser error occurs'
        if tok:
            self.error("unexpected token '%s'" % tok.value)
        else:
            self.error('unexpected EOF')

    def error(self, msg, lineno=0):
        'raise a SMVError error'
        lineno = lineno if lineno else self.lexer.lineno
        raise SMVError('%s:%d: error: %s' % (self.filename, lineno, msg))
